/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSFrequency_H_
#define	_FANSFrequency_H_


#include "asn_application.h"

/* Including external dependencies */
#include "FANSFrequencyhf.h"
#include "FANSFrequencyvhf.h"
#include "FANSFrequencyuhf.h"
#include "FANSFrequencysatchannel.h"
#include "constr_CHOICE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Dependencies */
typedef enum FANSFrequency_PR {
	FANSFrequency_PR_NOTHING,	/* No components present */
	FANSFrequency_PR_frequencyhf,
	FANSFrequency_PR_frequencyvhf,
	FANSFrequency_PR_frequencyuhf,
	FANSFrequency_PR_frequencysatchannel
} FANSFrequency_PR;

/* FANSFrequency */
typedef struct FANSFrequency {
	FANSFrequency_PR present;
	union FANSFrequency_u {
		FANSFrequencyhf_t	 frequencyhf;
		FANSFrequencyvhf_t	 frequencyvhf;
		FANSFrequencyuhf_t	 frequencyuhf;
		FANSFrequencysatchannel_t	 frequencysatchannel;
	} choice;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} FANSFrequency_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSFrequency;

#ifdef __cplusplus
}
#endif

#endif	/* _FANSFrequency_H_ */
#include "asn_internal.h"
