/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSPosition_H_
#define	_FANSPosition_H_


#include "asn_application.h"

/* Including external dependencies */
#include "FANSFixName.h"
#include "FANSNavaid.h"
#include "FANSAirport.h"
#include "FANSLatitudeLongitude.h"
#include "FANSPlaceBearingDistance.h"
#include "constr_CHOICE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Dependencies */
typedef enum FANSPosition_PR {
	FANSPosition_PR_NOTHING,	/* No components present */
	FANSPosition_PR_fixName,
	FANSPosition_PR_navaid,
	FANSPosition_PR_airport,
	FANSPosition_PR_latitudeLongitude,
	FANSPosition_PR_placeBearingDistance
} FANSPosition_PR;

/* FANSPosition */
typedef struct FANSPosition {
	FANSPosition_PR present;
	union FANSPosition_u {
		FANSFixName_t	 fixName;
		FANSNavaid_t	 navaid;
		FANSAirport_t	 airport;
		FANSLatitudeLongitude_t	 latitudeLongitude;
		FANSPlaceBearingDistance_t	 placeBearingDistance;
	} choice;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} FANSPosition_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSPosition;

#ifdef __cplusplus
}
#endif

#endif	/* _FANSPosition_H_ */
#include "asn_internal.h"
