/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSHoldatwaypoint_H_
#define	_FANSHoldatwaypoint_H_


#include "asn_application.h"

/* Including external dependencies */
#include "FANSPosition.h"
#include "FANSDirection.h"
#include "constr_SEQUENCE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Forward declarations */
struct FANSSpeed;
struct FANSATWAltitude;
struct FANSDegrees;
struct FANSTime;
struct FANSLegType;

/* FANSHoldatwaypoint */
typedef struct FANSHoldatwaypoint {
	FANSPosition_t	 position;
	struct FANSSpeed	*holdatwaypointspeedlow	/* OPTIONAL */;
	struct FANSATWAltitude	*aTWaltitude	/* OPTIONAL */;
	struct FANSSpeed	*holdatwaypointspeedhigh	/* OPTIONAL */;
	FANSDirection_t	*direction	/* OPTIONAL */;
	struct FANSDegrees	*degrees	/* OPTIONAL */;
	struct FANSTime	*eFCtime	/* OPTIONAL */;
	struct FANSLegType	*legType	/* OPTIONAL */;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} FANSHoldatwaypoint_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSHoldatwaypoint;

#ifdef __cplusplus
}
#endif

/* Referred external types */
#include "FANSHoldatwaypointSpeedLow.h"
#include "FANSATWAltitude.h"
#include "FANSHoldatwaypointSpeedHigh.h"
#include "FANSDegrees.h"
#include "FANSEFCtime.h"
#include "FANSLegType.h"

#endif	/* _FANSHoldatwaypoint_H_ */
#include "asn_internal.h"
