/*
 * jaricom.c
 *
 * Copyright (C) 1991-1998, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file holds place for arithmetic entropy codec tables.
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"

IJG_INT32 jaritab[1];	/* dummy table */
