/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSCOMNAVEquipmentStatus_H_
#define	_FANSCOMNAVEquipmentStatus_H_


#include "asn_application.h"

/* Including external dependencies */
#include "NativeEnumerated.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Dependencies */
typedef enum FANSCOMNAVEquipmentStatus {
	FANSCOMNAVEquipmentStatus_aloranA	= 0,
	FANSCOMNAVEquipmentStatus_cloranC	= 1,
	FANSCOMNAVEquipmentStatus_ddme	= 2,
	FANSCOMNAVEquipmentStatus_edecca	= 3,
	FANSCOMNAVEquipmentStatus_fadf	= 4,
	FANSCOMNAVEquipmentStatus_ggnss	= 5,
	FANSCOMNAVEquipmentStatus_hhfRTF	= 6,
	FANSCOMNAVEquipmentStatus_iinertialNavigation	= 7,
	FANSCOMNAVEquipmentStatus_lils	= 8,
	FANSCOMNAVEquipmentStatus_momega	= 9,
	FANSCOMNAVEquipmentStatus_ovor	= 10,
	FANSCOMNAVEquipmentStatus_pdoppler	= 11,
	FANSCOMNAVEquipmentStatus_rrnavRouteEquipment	= 12,
	FANSCOMNAVEquipmentStatus_ttacan	= 13,
	FANSCOMNAVEquipmentStatus_uuhfRTF	= 14,
	FANSCOMNAVEquipmentStatus_vvhfRTF	= 15
} e_FANSCOMNAVEquipmentStatus;

/* FANSCOMNAVEquipmentStatus */
typedef long	 FANSCOMNAVEquipmentStatus_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSCOMNAVEquipmentStatus;
asn_struct_free_f FANSCOMNAVEquipmentStatus_free;
asn_struct_print_f FANSCOMNAVEquipmentStatus_print;
asn_constr_check_f FANSCOMNAVEquipmentStatus_constraint;
ber_type_decoder_f FANSCOMNAVEquipmentStatus_decode_ber;
der_type_encoder_f FANSCOMNAVEquipmentStatus_encode_der;
xer_type_decoder_f FANSCOMNAVEquipmentStatus_decode_xer;
xer_type_encoder_f FANSCOMNAVEquipmentStatus_encode_xer;
per_type_decoder_f FANSCOMNAVEquipmentStatus_decode_uper;
per_type_encoder_f FANSCOMNAVEquipmentStatus_encode_uper;

#ifdef __cplusplus
}
#endif

#endif	/* _FANSCOMNAVEquipmentStatus_H_ */
#include "asn_internal.h"
