#ifndef _PSCOPE_
#define _PSCOPE_
#include "license.txt" 
#include <stdio.h>
#include <stdlib.h>

int PScopeShot(const char * filename, const char * title1, const char * title2, short * data, float samplerate, unsigned int numsamples );

#endif // _PSCOPE_
