/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSTrackDetail_H_
#define	_FANSTrackDetail_H_


#include "asn_application.h"

/* Including external dependencies */
#include "FANSTrackName.h"
#include "FANSLatitudeLongitudeSequence.h"
#include "constr_SEQUENCE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* FANSTrackDetail */
typedef struct FANSTrackDetail {
	FANSTrackName_t	 trackName;
	FANSLatitudeLongitudeSequence_t	 latitudeLongitude_seqOf;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} FANSTrackDetail_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSTrackDetail;

#ifdef __cplusplus
}
#endif

#endif	/* _FANSTrackDetail_H_ */
#include "asn_internal.h"
