/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSAltimeter_H_
#define	_FANSAltimeter_H_


#include "asn_application.h"

/* Including external dependencies */
#include "FANSAltimeterEnglish.h"
#include "FANSAltimeterMetric.h"
#include "constr_CHOICE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Dependencies */
typedef enum FANSAltimeter_PR {
	FANSAltimeter_PR_NOTHING,	/* No components present */
	FANSAltimeter_PR_altimeterEnglish,
	FANSAltimeter_PR_altimeterMetric
} FANSAltimeter_PR;

/* FANSAltimeter */
typedef struct FANSAltimeter {
	FANSAltimeter_PR present;
	union FANSAltimeter_u {
		FANSAltimeterEnglish_t	 altimeterEnglish;
		FANSAltimeterMetric_t	 altimeterMetric;
	} choice;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} FANSAltimeter_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSAltimeter;

#ifdef __cplusplus
}
#endif

#endif	/* _FANSAltimeter_H_ */
#include "asn_internal.h"
