//
//  opencl_header_private.h
//  libOpenCL_loader
//
//  Created by Guohui Wang on 4/23/18.
//  Copyright © 2019 Guohui Wang. All rights reserved.
//

#pragma once

#define CL_HPP_TARGET_OPENCL_VERSION 120
#define CL_HPP_MINIMUM_OPENCL_VERSION 110
#define CL_HPP_CL_1_2_DEFAULT_BUILD
#include "CL/cl.hpp"
