/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSRouteInformation_H_
#define	_FANSRouteInformation_H_


#include "asn_application.h"

/* Including external dependencies */
#include "FANSPublishedIdentifier.h"
#include "FANSLatitudeLongitude.h"
#include "FANSPlaceBearingPlaceBearing.h"
#include "FANSPlaceBearingDistance.h"
#include "FANSAirwayIdentifier.h"
#include "FANSTrackDetail.h"
#include "constr_CHOICE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Dependencies */
typedef enum FANSRouteInformation_PR {
	FANSRouteInformation_PR_NOTHING,	/* No components present */
	FANSRouteInformation_PR_publishedIdentifier,
	FANSRouteInformation_PR_latitudeLongitude,
	FANSRouteInformation_PR_placeBearingPlaceBearing,
	FANSRouteInformation_PR_placeBearingDistance,
	FANSRouteInformation_PR_airwayIdentifier,
	FANSRouteInformation_PR_trackDetail
} FANSRouteInformation_PR;

/* FANSRouteInformation */
typedef struct FANSRouteInformation {
	FANSRouteInformation_PR present;
	union FANSRouteInformation_u {
		FANSPublishedIdentifier_t	 publishedIdentifier;
		FANSLatitudeLongitude_t	 latitudeLongitude;
		FANSPlaceBearingPlaceBearing_t	 placeBearingPlaceBearing;
		FANSPlaceBearingDistance_t	 placeBearingDistance;
		FANSAirwayIdentifier_t	 airwayIdentifier;
		FANSTrackDetail_t	 trackDetail;
	} choice;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} FANSRouteInformation_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSRouteInformation;

#ifdef __cplusplus
}
#endif

#endif	/* _FANSRouteInformation_H_ */
#include "asn_internal.h"
