/*
 *  This file is a part of libacars
 *
 *  Copyright (c) 2018-2021 Tomasz Lemiech <szpajder@gmail.com>
 */
#ifndef _CONFIG_H
#define _CONFIG_H

#define WITH_ZLIB
// #define WITH_LIBXML2
/* #undef IS_BIG_ENDIAN */
#define HAVE_STRSEP
#define LFIND_NMEMB_SIZE_SIZE_T
/* #undef LFIND_NMEMB_SIZE_UINT */

#endif // !_CONFIG_H
