/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#include "FANSATCDownlinkMsgElementId.h"

static asn_per_constraints_t asn_PER_type_FANSATCDownlinkMsgElementId_constr_1 GCC_NOTUSED = {
	{ APC_CONSTRAINED,	 8,  8,  0,  128 }	/* (0..128) */,
	{ APC_UNCONSTRAINED,	-1, -1,  0,  0 },
	0, 0	/* No PER value map */
};
static asn_TYPE_member_t asn_MBR_FANSATCDownlinkMsgElementId_1[] = {
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM0NULL),
		(ASN_TAG_CLASS_CONTEXT | (0 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM0NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM1NULL),
		(ASN_TAG_CLASS_CONTEXT | (1 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM1NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM2NULL),
		(ASN_TAG_CLASS_CONTEXT | (2 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM2NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM3NULL),
		(ASN_TAG_CLASS_CONTEXT | (3 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM3NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM4NULL),
		(ASN_TAG_CLASS_CONTEXT | (4 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM4NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM5NULL),
		(ASN_TAG_CLASS_CONTEXT | (5 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM5NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM6Altitude),
		(ASN_TAG_CLASS_CONTEXT | (6 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM6Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM7AltitudeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (7 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM7AltitudeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM8Altitude),
		(ASN_TAG_CLASS_CONTEXT | (8 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM8Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM9Altitude),
		(ASN_TAG_CLASS_CONTEXT | (9 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM9Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM10Altitude),
		(ASN_TAG_CLASS_CONTEXT | (10 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM10Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM11PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (11 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM11PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM12PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (12 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM12PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM13TimeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (13 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM13TimeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM14TimeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (14 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM14TimeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM15DistanceOffsetDirection),
		(ASN_TAG_CLASS_CONTEXT | (15 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSDistanceOffsetDirection,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM15DistanceOffsetDirection"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM16PositionDistanceOffsetDirection),
		(ASN_TAG_CLASS_CONTEXT | (16 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionDistanceOffsetDirection,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM16PositionDistanceOffsetDirection"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM17TimeDistanceOffsetDirection),
		(ASN_TAG_CLASS_CONTEXT | (17 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeDistanceOffsetDirection,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM17TimeDistanceOffsetDirection"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM18Speed),
		(ASN_TAG_CLASS_CONTEXT | (18 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM18Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM19SpeedSpeed),
		(ASN_TAG_CLASS_CONTEXT | (19 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSSpeedSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM19SpeedSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM20NULL),
		(ASN_TAG_CLASS_CONTEXT | (20 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM20NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM21Frequency),
		(ASN_TAG_CLASS_CONTEXT | (21 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSFrequency,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM21Frequency"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM22Position),
		(ASN_TAG_CLASS_CONTEXT | (22 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM22Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM23ProcedureName),
		(ASN_TAG_CLASS_CONTEXT | (23 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSProcedureName,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM23ProcedureName"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM24RouteClearance),
		(ASN_TAG_CLASS_CONTEXT | (24 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSRouteClearance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM24RouteClearance"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM25NULL),
		(ASN_TAG_CLASS_CONTEXT | (25 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM25NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM26PositionRouteClearance),
		(ASN_TAG_CLASS_CONTEXT | (26 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionRouteClearance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM26PositionRouteClearance"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM27DistanceOffsetDirection),
		(ASN_TAG_CLASS_CONTEXT | (27 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSDistanceOffsetDirection,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM27DistanceOffsetDirection"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM28Altitude),
		(ASN_TAG_CLASS_CONTEXT | (28 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM28Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM29Altitude),
		(ASN_TAG_CLASS_CONTEXT | (29 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM29Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM30Altitude),
		(ASN_TAG_CLASS_CONTEXT | (30 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM30Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM31Position),
		(ASN_TAG_CLASS_CONTEXT | (31 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM31Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM32Altitude),
		(ASN_TAG_CLASS_CONTEXT | (32 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM32Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM33Position),
		(ASN_TAG_CLASS_CONTEXT | (33 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM33Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM34Speed),
		(ASN_TAG_CLASS_CONTEXT | (34 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM34Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM35Degrees),
		(ASN_TAG_CLASS_CONTEXT | (35 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSDegrees,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM35Degrees"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM36Degrees),
		(ASN_TAG_CLASS_CONTEXT | (36 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSDegrees,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM36Degrees"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM37Altitude),
		(ASN_TAG_CLASS_CONTEXT | (37 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM37Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM38Altitude),
		(ASN_TAG_CLASS_CONTEXT | (38 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM38Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM39Speed),
		(ASN_TAG_CLASS_CONTEXT | (39 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM39Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM40RouteClearance),
		(ASN_TAG_CLASS_CONTEXT | (40 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSRouteClearance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM40RouteClearance"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM41NULL),
		(ASN_TAG_CLASS_CONTEXT | (41 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM41NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM42Position),
		(ASN_TAG_CLASS_CONTEXT | (42 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM42Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM43Time),
		(ASN_TAG_CLASS_CONTEXT | (43 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM43Time"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM44Position),
		(ASN_TAG_CLASS_CONTEXT | (44 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM44Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM45Position),
		(ASN_TAG_CLASS_CONTEXT | (45 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM45Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM46Time),
		(ASN_TAG_CLASS_CONTEXT | (46 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM46Time"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM47BeaconCode),
		(ASN_TAG_CLASS_CONTEXT | (47 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSBeaconCode,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM47BeaconCode"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM48PositionReport),
		(ASN_TAG_CLASS_CONTEXT | (48 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionReport,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM48PositionReport"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM49Speed),
		(ASN_TAG_CLASS_CONTEXT | (49 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM49Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM50SpeedSpeed),
		(ASN_TAG_CLASS_CONTEXT | (50 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSSpeedSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM50SpeedSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM51NULL),
		(ASN_TAG_CLASS_CONTEXT | (51 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM51NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM52NULL),
		(ASN_TAG_CLASS_CONTEXT | (52 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM52NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM53NULL),
		(ASN_TAG_CLASS_CONTEXT | (53 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM53NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM54Altitude),
		(ASN_TAG_CLASS_CONTEXT | (54 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM54Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM55NULL),
		(ASN_TAG_CLASS_CONTEXT | (55 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM55NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM56NULL),
		(ASN_TAG_CLASS_CONTEXT | (56 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM56NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM57RemainingFuelRemainingSouls),
		(ASN_TAG_CLASS_CONTEXT | (57 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSRemainingFuelRemainingSouls,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM57RemainingFuelRemainingSouls"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM58NULL),
		(ASN_TAG_CLASS_CONTEXT | (58 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM58NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM59PositionRouteClearance),
		(ASN_TAG_CLASS_CONTEXT | (59 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionRouteClearance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM59PositionRouteClearance"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM60DistanceOffsetDirection),
		(ASN_TAG_CLASS_CONTEXT | (60 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSDistanceOffsetDirection,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM60DistanceOffsetDirection"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM61Altitude),
		(ASN_TAG_CLASS_CONTEXT | (61 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM61Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM62ErrorInformation),
		(ASN_TAG_CLASS_CONTEXT | (62 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSErrorInformation,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM62ErrorInformation"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM63NULL),
		(ASN_TAG_CLASS_CONTEXT | (63 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM63NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM64ICAOfacilitydesignation),
		(ASN_TAG_CLASS_CONTEXT | (64 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSICAOfacilityDesignation,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM64ICAOfacilitydesignation"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM65NULL),
		(ASN_TAG_CLASS_CONTEXT | (65 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM65NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM66NULL),
		(ASN_TAG_CLASS_CONTEXT | (66 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM66NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM67FreeText),
		(ASN_TAG_CLASS_CONTEXT | (67 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSFreeText,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM67FreeText"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM68FreeText),
		(ASN_TAG_CLASS_CONTEXT | (68 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSFreeText,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM68FreeText"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM69NULL),
		(ASN_TAG_CLASS_CONTEXT | (69 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM69NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM70Degrees),
		(ASN_TAG_CLASS_CONTEXT | (70 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSDegrees,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM70Degrees"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM71Degrees),
		(ASN_TAG_CLASS_CONTEXT | (71 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSDegrees,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM71Degrees"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM72Altitude),
		(ASN_TAG_CLASS_CONTEXT | (72 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM72Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM73VersionNumber),
		(ASN_TAG_CLASS_CONTEXT | (73 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSVersionNumber,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM73VersionNumber"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM74NULL),
		(ASN_TAG_CLASS_CONTEXT | (74 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM74NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM75NULL),
		(ASN_TAG_CLASS_CONTEXT | (75 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM75NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM76AltitudeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (76 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM76AltitudeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM77AltitudeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (77 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM77AltitudeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM78TimeDistanceToFromPosition),
		(ASN_TAG_CLASS_CONTEXT | (78 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeDistanceToFromPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM78TimeDistanceToFromPosition"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM79ATISCode),
		(ASN_TAG_CLASS_CONTEXT | (79 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSATISCode,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM79ATISCode"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM80DistanceOffsetDirection),
		(ASN_TAG_CLASS_CONTEXT | (80 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSDistanceOffsetDirection,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM80DistanceOffsetDirection"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM81NULL),
		(ASN_TAG_CLASS_CONTEXT | (81 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM81NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM82NULL),
		(ASN_TAG_CLASS_CONTEXT | (82 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM82NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM83NULL),
		(ASN_TAG_CLASS_CONTEXT | (83 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM83NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM84NULL),
		(ASN_TAG_CLASS_CONTEXT | (84 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM84NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM85NULL),
		(ASN_TAG_CLASS_CONTEXT | (85 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM85NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM86NULL),
		(ASN_TAG_CLASS_CONTEXT | (86 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM86NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM87NULL),
		(ASN_TAG_CLASS_CONTEXT | (87 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM87NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM88NULL),
		(ASN_TAG_CLASS_CONTEXT | (88 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM88NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM89NULL),
		(ASN_TAG_CLASS_CONTEXT | (89 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM89NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM90NULL),
		(ASN_TAG_CLASS_CONTEXT | (90 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM90NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM91NULL),
		(ASN_TAG_CLASS_CONTEXT | (91 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM91NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM92NULL),
		(ASN_TAG_CLASS_CONTEXT | (92 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM92NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM93NULL),
		(ASN_TAG_CLASS_CONTEXT | (93 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM93NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM94NULL),
		(ASN_TAG_CLASS_CONTEXT | (94 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM94NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM95NULL),
		(ASN_TAG_CLASS_CONTEXT | (95 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM95NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM96NULL),
		(ASN_TAG_CLASS_CONTEXT | (96 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM96NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM97NULL),
		(ASN_TAG_CLASS_CONTEXT | (97 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM97NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM98NULL),
		(ASN_TAG_CLASS_CONTEXT | (98 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM98NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM99NULL),
		(ASN_TAG_CLASS_CONTEXT | (99 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM99NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM100NULL),
		(ASN_TAG_CLASS_CONTEXT | (100 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM100NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM101NULL),
		(ASN_TAG_CLASS_CONTEXT | (101 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM101NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM102NULL),
		(ASN_TAG_CLASS_CONTEXT | (102 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM102NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM103NULL),
		(ASN_TAG_CLASS_CONTEXT | (103 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM103NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM104NULL),
		(ASN_TAG_CLASS_CONTEXT | (104 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM104NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM105NULL),
		(ASN_TAG_CLASS_CONTEXT | (105 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM105NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM106NULL),
		(ASN_TAG_CLASS_CONTEXT | (106 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM106NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM107NULL),
		(ASN_TAG_CLASS_CONTEXT | (107 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM107NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM108NULL),
		(ASN_TAG_CLASS_CONTEXT | (108 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM108NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM109NULL),
		(ASN_TAG_CLASS_CONTEXT | (109 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM109NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM110NULL),
		(ASN_TAG_CLASS_CONTEXT | (110 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM110NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM111NULL),
		(ASN_TAG_CLASS_CONTEXT | (111 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM111NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM112NULL),
		(ASN_TAG_CLASS_CONTEXT | (112 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM112NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM113NULL),
		(ASN_TAG_CLASS_CONTEXT | (113 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM113NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM114NULL),
		(ASN_TAG_CLASS_CONTEXT | (114 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM114NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM115NULL),
		(ASN_TAG_CLASS_CONTEXT | (115 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM115NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM116NULL),
		(ASN_TAG_CLASS_CONTEXT | (116 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM116NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM117NULL),
		(ASN_TAG_CLASS_CONTEXT | (117 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM117NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM118NULL),
		(ASN_TAG_CLASS_CONTEXT | (118 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM118NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM119NULL),
		(ASN_TAG_CLASS_CONTEXT | (119 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM119NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM120NULL),
		(ASN_TAG_CLASS_CONTEXT | (120 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM120NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM121NULL),
		(ASN_TAG_CLASS_CONTEXT | (121 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM121NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM122NULL),
		(ASN_TAG_CLASS_CONTEXT | (122 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM122NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM123NULL),
		(ASN_TAG_CLASS_CONTEXT | (123 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM123NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM124NULL),
		(ASN_TAG_CLASS_CONTEXT | (124 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM124NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM125NULL),
		(ASN_TAG_CLASS_CONTEXT | (125 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM125NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM126NULL),
		(ASN_TAG_CLASS_CONTEXT | (126 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM126NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM127NULL),
		(ASN_TAG_CLASS_CONTEXT | (127 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM127NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCDownlinkMsgElementId, choice.dM128NULL),
		(ASN_TAG_CLASS_CONTEXT | (128 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"dM128NULL"
		},
};
static const asn_TYPE_tag2member_t asn_MAP_FANSATCDownlinkMsgElementId_tag2el_1[] = {
    { (ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0 }, /* dM0NULL */
    { (ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0 }, /* dM1NULL */
    { (ASN_TAG_CLASS_CONTEXT | (2 << 2)), 2, 0, 0 }, /* dM2NULL */
    { (ASN_TAG_CLASS_CONTEXT | (3 << 2)), 3, 0, 0 }, /* dM3NULL */
    { (ASN_TAG_CLASS_CONTEXT | (4 << 2)), 4, 0, 0 }, /* dM4NULL */
    { (ASN_TAG_CLASS_CONTEXT | (5 << 2)), 5, 0, 0 }, /* dM5NULL */
    { (ASN_TAG_CLASS_CONTEXT | (6 << 2)), 6, 0, 0 }, /* dM6Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (7 << 2)), 7, 0, 0 }, /* dM7AltitudeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (8 << 2)), 8, 0, 0 }, /* dM8Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (9 << 2)), 9, 0, 0 }, /* dM9Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (10 << 2)), 10, 0, 0 }, /* dM10Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (11 << 2)), 11, 0, 0 }, /* dM11PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (12 << 2)), 12, 0, 0 }, /* dM12PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (13 << 2)), 13, 0, 0 }, /* dM13TimeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (14 << 2)), 14, 0, 0 }, /* dM14TimeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (15 << 2)), 15, 0, 0 }, /* dM15DistanceOffsetDirection */
    { (ASN_TAG_CLASS_CONTEXT | (16 << 2)), 16, 0, 0 }, /* dM16PositionDistanceOffsetDirection */
    { (ASN_TAG_CLASS_CONTEXT | (17 << 2)), 17, 0, 0 }, /* dM17TimeDistanceOffsetDirection */
    { (ASN_TAG_CLASS_CONTEXT | (18 << 2)), 18, 0, 0 }, /* dM18Speed */
    { (ASN_TAG_CLASS_CONTEXT | (19 << 2)), 19, 0, 0 }, /* dM19SpeedSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (20 << 2)), 20, 0, 0 }, /* dM20NULL */
    { (ASN_TAG_CLASS_CONTEXT | (21 << 2)), 21, 0, 0 }, /* dM21Frequency */
    { (ASN_TAG_CLASS_CONTEXT | (22 << 2)), 22, 0, 0 }, /* dM22Position */
    { (ASN_TAG_CLASS_CONTEXT | (23 << 2)), 23, 0, 0 }, /* dM23ProcedureName */
    { (ASN_TAG_CLASS_CONTEXT | (24 << 2)), 24, 0, 0 }, /* dM24RouteClearance */
    { (ASN_TAG_CLASS_CONTEXT | (25 << 2)), 25, 0, 0 }, /* dM25NULL */
    { (ASN_TAG_CLASS_CONTEXT | (26 << 2)), 26, 0, 0 }, /* dM26PositionRouteClearance */
    { (ASN_TAG_CLASS_CONTEXT | (27 << 2)), 27, 0, 0 }, /* dM27DistanceOffsetDirection */
    { (ASN_TAG_CLASS_CONTEXT | (28 << 2)), 28, 0, 0 }, /* dM28Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (29 << 2)), 29, 0, 0 }, /* dM29Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (30 << 2)), 30, 0, 0 }, /* dM30Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (31 << 2)), 31, 0, 0 }, /* dM31Position */
    { (ASN_TAG_CLASS_CONTEXT | (32 << 2)), 32, 0, 0 }, /* dM32Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (33 << 2)), 33, 0, 0 }, /* dM33Position */
    { (ASN_TAG_CLASS_CONTEXT | (34 << 2)), 34, 0, 0 }, /* dM34Speed */
    { (ASN_TAG_CLASS_CONTEXT | (35 << 2)), 35, 0, 0 }, /* dM35Degrees */
    { (ASN_TAG_CLASS_CONTEXT | (36 << 2)), 36, 0, 0 }, /* dM36Degrees */
    { (ASN_TAG_CLASS_CONTEXT | (37 << 2)), 37, 0, 0 }, /* dM37Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (38 << 2)), 38, 0, 0 }, /* dM38Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (39 << 2)), 39, 0, 0 }, /* dM39Speed */
    { (ASN_TAG_CLASS_CONTEXT | (40 << 2)), 40, 0, 0 }, /* dM40RouteClearance */
    { (ASN_TAG_CLASS_CONTEXT | (41 << 2)), 41, 0, 0 }, /* dM41NULL */
    { (ASN_TAG_CLASS_CONTEXT | (42 << 2)), 42, 0, 0 }, /* dM42Position */
    { (ASN_TAG_CLASS_CONTEXT | (43 << 2)), 43, 0, 0 }, /* dM43Time */
    { (ASN_TAG_CLASS_CONTEXT | (44 << 2)), 44, 0, 0 }, /* dM44Position */
    { (ASN_TAG_CLASS_CONTEXT | (45 << 2)), 45, 0, 0 }, /* dM45Position */
    { (ASN_TAG_CLASS_CONTEXT | (46 << 2)), 46, 0, 0 }, /* dM46Time */
    { (ASN_TAG_CLASS_CONTEXT | (47 << 2)), 47, 0, 0 }, /* dM47BeaconCode */
    { (ASN_TAG_CLASS_CONTEXT | (48 << 2)), 48, 0, 0 }, /* dM48PositionReport */
    { (ASN_TAG_CLASS_CONTEXT | (49 << 2)), 49, 0, 0 }, /* dM49Speed */
    { (ASN_TAG_CLASS_CONTEXT | (50 << 2)), 50, 0, 0 }, /* dM50SpeedSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (51 << 2)), 51, 0, 0 }, /* dM51NULL */
    { (ASN_TAG_CLASS_CONTEXT | (52 << 2)), 52, 0, 0 }, /* dM52NULL */
    { (ASN_TAG_CLASS_CONTEXT | (53 << 2)), 53, 0, 0 }, /* dM53NULL */
    { (ASN_TAG_CLASS_CONTEXT | (54 << 2)), 54, 0, 0 }, /* dM54Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (55 << 2)), 55, 0, 0 }, /* dM55NULL */
    { (ASN_TAG_CLASS_CONTEXT | (56 << 2)), 56, 0, 0 }, /* dM56NULL */
    { (ASN_TAG_CLASS_CONTEXT | (57 << 2)), 57, 0, 0 }, /* dM57RemainingFuelRemainingSouls */
    { (ASN_TAG_CLASS_CONTEXT | (58 << 2)), 58, 0, 0 }, /* dM58NULL */
    { (ASN_TAG_CLASS_CONTEXT | (59 << 2)), 59, 0, 0 }, /* dM59PositionRouteClearance */
    { (ASN_TAG_CLASS_CONTEXT | (60 << 2)), 60, 0, 0 }, /* dM60DistanceOffsetDirection */
    { (ASN_TAG_CLASS_CONTEXT | (61 << 2)), 61, 0, 0 }, /* dM61Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (62 << 2)), 62, 0, 0 }, /* dM62ErrorInformation */
    { (ASN_TAG_CLASS_CONTEXT | (63 << 2)), 63, 0, 0 }, /* dM63NULL */
    { (ASN_TAG_CLASS_CONTEXT | (64 << 2)), 64, 0, 0 }, /* dM64ICAOfacilitydesignation */
    { (ASN_TAG_CLASS_CONTEXT | (65 << 2)), 65, 0, 0 }, /* dM65NULL */
    { (ASN_TAG_CLASS_CONTEXT | (66 << 2)), 66, 0, 0 }, /* dM66NULL */
    { (ASN_TAG_CLASS_CONTEXT | (67 << 2)), 67, 0, 0 }, /* dM67FreeText */
    { (ASN_TAG_CLASS_CONTEXT | (68 << 2)), 68, 0, 0 }, /* dM68FreeText */
    { (ASN_TAG_CLASS_CONTEXT | (69 << 2)), 69, 0, 0 }, /* dM69NULL */
    { (ASN_TAG_CLASS_CONTEXT | (70 << 2)), 70, 0, 0 }, /* dM70Degrees */
    { (ASN_TAG_CLASS_CONTEXT | (71 << 2)), 71, 0, 0 }, /* dM71Degrees */
    { (ASN_TAG_CLASS_CONTEXT | (72 << 2)), 72, 0, 0 }, /* dM72Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (73 << 2)), 73, 0, 0 }, /* dM73VersionNumber */
    { (ASN_TAG_CLASS_CONTEXT | (74 << 2)), 74, 0, 0 }, /* dM74NULL */
    { (ASN_TAG_CLASS_CONTEXT | (75 << 2)), 75, 0, 0 }, /* dM75NULL */
    { (ASN_TAG_CLASS_CONTEXT | (76 << 2)), 76, 0, 0 }, /* dM76AltitudeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (77 << 2)), 77, 0, 0 }, /* dM77AltitudeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (78 << 2)), 78, 0, 0 }, /* dM78TimeDistanceToFromPosition */
    { (ASN_TAG_CLASS_CONTEXT | (79 << 2)), 79, 0, 0 }, /* dM79ATISCode */
    { (ASN_TAG_CLASS_CONTEXT | (80 << 2)), 80, 0, 0 }, /* dM80DistanceOffsetDirection */
    { (ASN_TAG_CLASS_CONTEXT | (81 << 2)), 81, 0, 0 }, /* dM81NULL */
    { (ASN_TAG_CLASS_CONTEXT | (82 << 2)), 82, 0, 0 }, /* dM82NULL */
    { (ASN_TAG_CLASS_CONTEXT | (83 << 2)), 83, 0, 0 }, /* dM83NULL */
    { (ASN_TAG_CLASS_CONTEXT | (84 << 2)), 84, 0, 0 }, /* dM84NULL */
    { (ASN_TAG_CLASS_CONTEXT | (85 << 2)), 85, 0, 0 }, /* dM85NULL */
    { (ASN_TAG_CLASS_CONTEXT | (86 << 2)), 86, 0, 0 }, /* dM86NULL */
    { (ASN_TAG_CLASS_CONTEXT | (87 << 2)), 87, 0, 0 }, /* dM87NULL */
    { (ASN_TAG_CLASS_CONTEXT | (88 << 2)), 88, 0, 0 }, /* dM88NULL */
    { (ASN_TAG_CLASS_CONTEXT | (89 << 2)), 89, 0, 0 }, /* dM89NULL */
    { (ASN_TAG_CLASS_CONTEXT | (90 << 2)), 90, 0, 0 }, /* dM90NULL */
    { (ASN_TAG_CLASS_CONTEXT | (91 << 2)), 91, 0, 0 }, /* dM91NULL */
    { (ASN_TAG_CLASS_CONTEXT | (92 << 2)), 92, 0, 0 }, /* dM92NULL */
    { (ASN_TAG_CLASS_CONTEXT | (93 << 2)), 93, 0, 0 }, /* dM93NULL */
    { (ASN_TAG_CLASS_CONTEXT | (94 << 2)), 94, 0, 0 }, /* dM94NULL */
    { (ASN_TAG_CLASS_CONTEXT | (95 << 2)), 95, 0, 0 }, /* dM95NULL */
    { (ASN_TAG_CLASS_CONTEXT | (96 << 2)), 96, 0, 0 }, /* dM96NULL */
    { (ASN_TAG_CLASS_CONTEXT | (97 << 2)), 97, 0, 0 }, /* dM97NULL */
    { (ASN_TAG_CLASS_CONTEXT | (98 << 2)), 98, 0, 0 }, /* dM98NULL */
    { (ASN_TAG_CLASS_CONTEXT | (99 << 2)), 99, 0, 0 }, /* dM99NULL */
    { (ASN_TAG_CLASS_CONTEXT | (100 << 2)), 100, 0, 0 }, /* dM100NULL */
    { (ASN_TAG_CLASS_CONTEXT | (101 << 2)), 101, 0, 0 }, /* dM101NULL */
    { (ASN_TAG_CLASS_CONTEXT | (102 << 2)), 102, 0, 0 }, /* dM102NULL */
    { (ASN_TAG_CLASS_CONTEXT | (103 << 2)), 103, 0, 0 }, /* dM103NULL */
    { (ASN_TAG_CLASS_CONTEXT | (104 << 2)), 104, 0, 0 }, /* dM104NULL */
    { (ASN_TAG_CLASS_CONTEXT | (105 << 2)), 105, 0, 0 }, /* dM105NULL */
    { (ASN_TAG_CLASS_CONTEXT | (106 << 2)), 106, 0, 0 }, /* dM106NULL */
    { (ASN_TAG_CLASS_CONTEXT | (107 << 2)), 107, 0, 0 }, /* dM107NULL */
    { (ASN_TAG_CLASS_CONTEXT | (108 << 2)), 108, 0, 0 }, /* dM108NULL */
    { (ASN_TAG_CLASS_CONTEXT | (109 << 2)), 109, 0, 0 }, /* dM109NULL */
    { (ASN_TAG_CLASS_CONTEXT | (110 << 2)), 110, 0, 0 }, /* dM110NULL */
    { (ASN_TAG_CLASS_CONTEXT | (111 << 2)), 111, 0, 0 }, /* dM111NULL */
    { (ASN_TAG_CLASS_CONTEXT | (112 << 2)), 112, 0, 0 }, /* dM112NULL */
    { (ASN_TAG_CLASS_CONTEXT | (113 << 2)), 113, 0, 0 }, /* dM113NULL */
    { (ASN_TAG_CLASS_CONTEXT | (114 << 2)), 114, 0, 0 }, /* dM114NULL */
    { (ASN_TAG_CLASS_CONTEXT | (115 << 2)), 115, 0, 0 }, /* dM115NULL */
    { (ASN_TAG_CLASS_CONTEXT | (116 << 2)), 116, 0, 0 }, /* dM116NULL */
    { (ASN_TAG_CLASS_CONTEXT | (117 << 2)), 117, 0, 0 }, /* dM117NULL */
    { (ASN_TAG_CLASS_CONTEXT | (118 << 2)), 118, 0, 0 }, /* dM118NULL */
    { (ASN_TAG_CLASS_CONTEXT | (119 << 2)), 119, 0, 0 }, /* dM119NULL */
    { (ASN_TAG_CLASS_CONTEXT | (120 << 2)), 120, 0, 0 }, /* dM120NULL */
    { (ASN_TAG_CLASS_CONTEXT | (121 << 2)), 121, 0, 0 }, /* dM121NULL */
    { (ASN_TAG_CLASS_CONTEXT | (122 << 2)), 122, 0, 0 }, /* dM122NULL */
    { (ASN_TAG_CLASS_CONTEXT | (123 << 2)), 123, 0, 0 }, /* dM123NULL */
    { (ASN_TAG_CLASS_CONTEXT | (124 << 2)), 124, 0, 0 }, /* dM124NULL */
    { (ASN_TAG_CLASS_CONTEXT | (125 << 2)), 125, 0, 0 }, /* dM125NULL */
    { (ASN_TAG_CLASS_CONTEXT | (126 << 2)), 126, 0, 0 }, /* dM126NULL */
    { (ASN_TAG_CLASS_CONTEXT | (127 << 2)), 127, 0, 0 }, /* dM127NULL */
    { (ASN_TAG_CLASS_CONTEXT | (128 << 2)), 128, 0, 0 } /* dM128NULL */
};
static asn_CHOICE_specifics_t asn_SPC_FANSATCDownlinkMsgElementId_specs_1 = {
	sizeof(struct FANSATCDownlinkMsgElementId),
	offsetof(struct FANSATCDownlinkMsgElementId, _asn_ctx),
	offsetof(struct FANSATCDownlinkMsgElementId, present),
	sizeof(((struct FANSATCDownlinkMsgElementId *)0)->present),
	asn_MAP_FANSATCDownlinkMsgElementId_tag2el_1,
	129,	/* Count of tags in the map */
	0,
	-1	/* Extensions start */
};
asn_TYPE_descriptor_t asn_DEF_FANSATCDownlinkMsgElementId = {
	"FANSATCDownlinkMsgElementId",
	"FANSATCDownlinkMsgElementId",
	CHOICE_free,
	CHOICE_print,
	CHOICE_constraint,
	CHOICE_decode_ber,
	CHOICE_encode_der,
	CHOICE_decode_xer,
	CHOICE_encode_xer,
	CHOICE_decode_uper,
	CHOICE_encode_uper,
	CHOICE_outmost_tag,
	0,	/* No effective tags (pointer) */
	0,	/* No effective tags (count) */
	0,	/* No tags (pointer) */
	0,	/* No tags (count) */
	&asn_PER_type_FANSATCDownlinkMsgElementId_constr_1,
	asn_MBR_FANSATCDownlinkMsgElementId_1,
	129,	/* Elements count */
	&asn_SPC_FANSATCDownlinkMsgElementId_specs_1	/* Additional specs */
};

