/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#include "FANSPositionReport.h"

static asn_TYPE_member_t asn_MBR_FANSPositionReport_1[] = {
	{ ATF_NOFLAGS, 0, offsetof(struct FANSPositionReport, positioncurrent),
		(ASN_TAG_CLASS_CONTEXT | (0 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPositionCurrent,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"positioncurrent"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSPositionReport, timeatpositioncurrent),
		(ASN_TAG_CLASS_CONTEXT | (1 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeAtPositionCurrent,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"timeatpositioncurrent"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSPositionReport, altitude),
		(ASN_TAG_CLASS_CONTEXT | (2 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"altitude"
		},
	{ ATF_POINTER, 19, offsetof(struct FANSPositionReport, fixnext),
		(ASN_TAG_CLASS_CONTEXT | (3 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSFixNext,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"fixnext"
		},
	{ ATF_POINTER, 18, offsetof(struct FANSPositionReport, timeetaatfixnext),
		(ASN_TAG_CLASS_CONTEXT | (4 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeEtaAtFixNext,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"timeetaatfixnext"
		},
	{ ATF_POINTER, 17, offsetof(struct FANSPositionReport, fixnextplusone),
		(ASN_TAG_CLASS_CONTEXT | (5 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSFixNextPlusOne,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"fixnextplusone"
		},
	{ ATF_POINTER, 16, offsetof(struct FANSPositionReport, timeetadestination),
		(ASN_TAG_CLASS_CONTEXT | (6 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeEtaDestination,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"timeetadestination"
		},
	{ ATF_POINTER, 15, offsetof(struct FANSPositionReport, remainingFuel),
		(ASN_TAG_CLASS_CONTEXT | (7 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSRemainingFuel,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"remainingFuel"
		},
	{ ATF_POINTER, 14, offsetof(struct FANSPositionReport, temperature),
		(ASN_TAG_CLASS_CONTEXT | (8 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSTemperature,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"temperature"
		},
	{ ATF_POINTER, 13, offsetof(struct FANSPositionReport, winds),
		(ASN_TAG_CLASS_CONTEXT | (9 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSWinds,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"winds"
		},
	{ ATF_POINTER, 12, offsetof(struct FANSPositionReport, turbulence),
		(ASN_TAG_CLASS_CONTEXT | (10 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTurbulence,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"turbulence"
		},
	{ ATF_POINTER, 11, offsetof(struct FANSPositionReport, icing),
		(ASN_TAG_CLASS_CONTEXT | (11 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSIcing,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"icing"
		},
	{ ATF_POINTER, 10, offsetof(struct FANSPositionReport, speed),
		(ASN_TAG_CLASS_CONTEXT | (12 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"speed"
		},
	{ ATF_POINTER, 9, offsetof(struct FANSPositionReport, speedground),
		(ASN_TAG_CLASS_CONTEXT | (13 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSSpeedGround,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"speedground"
		},
	{ ATF_POINTER, 8, offsetof(struct FANSPositionReport, verticalChange),
		(ASN_TAG_CLASS_CONTEXT | (14 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSVerticalChange,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"verticalChange"
		},
	{ ATF_POINTER, 7, offsetof(struct FANSPositionReport, trackAngle),
		(ASN_TAG_CLASS_CONTEXT | (15 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSTrackAngle,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"trackAngle"
		},
	{ ATF_POINTER, 6, offsetof(struct FANSPositionReport, trueHeading),
		(ASN_TAG_CLASS_CONTEXT | (16 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSTrueheading,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"trueHeading"
		},
	{ ATF_POINTER, 5, offsetof(struct FANSPositionReport, distance),
		(ASN_TAG_CLASS_CONTEXT | (17 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSDistance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"distance"
		},
	{ ATF_POINTER, 4, offsetof(struct FANSPositionReport, supplementaryInformation),
		(ASN_TAG_CLASS_CONTEXT | (18 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSSupplementaryInformation,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"supplementaryInformation"
		},
	{ ATF_POINTER, 3, offsetof(struct FANSPositionReport, reportedWaypointPosition),
		(ASN_TAG_CLASS_CONTEXT | (19 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSReportedWaypointPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"reportedWaypointPosition"
		},
	{ ATF_POINTER, 2, offsetof(struct FANSPositionReport, reportedWaypointTime),
		(ASN_TAG_CLASS_CONTEXT | (20 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSReportedWaypointTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"reportedWaypointTime"
		},
	{ ATF_POINTER, 1, offsetof(struct FANSPositionReport, reportedWaypointAltitude),
		(ASN_TAG_CLASS_CONTEXT | (21 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSReportedWaypointAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"reportedWaypointAltitude"
		},
};
static const int asn_MAP_FANSPositionReport_oms_1[] = { 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21 };
static const ber_tlv_tag_t asn_DEF_FANSPositionReport_tags_1[] = {
	(ASN_TAG_CLASS_UNIVERSAL | (16 << 2))
};
static const asn_TYPE_tag2member_t asn_MAP_FANSPositionReport_tag2el_1[] = {
    { (ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0 }, /* positioncurrent */
    { (ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0 }, /* timeatpositioncurrent */
    { (ASN_TAG_CLASS_CONTEXT | (2 << 2)), 2, 0, 0 }, /* altitude */
    { (ASN_TAG_CLASS_CONTEXT | (3 << 2)), 3, 0, 0 }, /* fixnext */
    { (ASN_TAG_CLASS_CONTEXT | (4 << 2)), 4, 0, 0 }, /* timeetaatfixnext */
    { (ASN_TAG_CLASS_CONTEXT | (5 << 2)), 5, 0, 0 }, /* fixnextplusone */
    { (ASN_TAG_CLASS_CONTEXT | (6 << 2)), 6, 0, 0 }, /* timeetadestination */
    { (ASN_TAG_CLASS_CONTEXT | (7 << 2)), 7, 0, 0 }, /* remainingFuel */
    { (ASN_TAG_CLASS_CONTEXT | (8 << 2)), 8, 0, 0 }, /* temperature */
    { (ASN_TAG_CLASS_CONTEXT | (9 << 2)), 9, 0, 0 }, /* winds */
    { (ASN_TAG_CLASS_CONTEXT | (10 << 2)), 10, 0, 0 }, /* turbulence */
    { (ASN_TAG_CLASS_CONTEXT | (11 << 2)), 11, 0, 0 }, /* icing */
    { (ASN_TAG_CLASS_CONTEXT | (12 << 2)), 12, 0, 0 }, /* speed */
    { (ASN_TAG_CLASS_CONTEXT | (13 << 2)), 13, 0, 0 }, /* speedground */
    { (ASN_TAG_CLASS_CONTEXT | (14 << 2)), 14, 0, 0 }, /* verticalChange */
    { (ASN_TAG_CLASS_CONTEXT | (15 << 2)), 15, 0, 0 }, /* trackAngle */
    { (ASN_TAG_CLASS_CONTEXT | (16 << 2)), 16, 0, 0 }, /* trueHeading */
    { (ASN_TAG_CLASS_CONTEXT | (17 << 2)), 17, 0, 0 }, /* distance */
    { (ASN_TAG_CLASS_CONTEXT | (18 << 2)), 18, 0, 0 }, /* supplementaryInformation */
    { (ASN_TAG_CLASS_CONTEXT | (19 << 2)), 19, 0, 0 }, /* reportedWaypointPosition */
    { (ASN_TAG_CLASS_CONTEXT | (20 << 2)), 20, 0, 0 }, /* reportedWaypointTime */
    { (ASN_TAG_CLASS_CONTEXT | (21 << 2)), 21, 0, 0 } /* reportedWaypointAltitude */
};
static asn_SEQUENCE_specifics_t asn_SPC_FANSPositionReport_specs_1 = {
	sizeof(struct FANSPositionReport),
	offsetof(struct FANSPositionReport, _asn_ctx),
	asn_MAP_FANSPositionReport_tag2el_1,
	22,	/* Count of tags in the map */
	asn_MAP_FANSPositionReport_oms_1,	/* Optional members */
	19, 0,	/* Root/Additions */
	-1,	/* Start extensions */
	-1	/* Stop extensions */
};
asn_TYPE_descriptor_t asn_DEF_FANSPositionReport = {
	"FANSPositionReport",
	"FANSPositionReport",
	SEQUENCE_free,
	SEQUENCE_print,
	SEQUENCE_constraint,
	SEQUENCE_decode_ber,
	SEQUENCE_encode_der,
	SEQUENCE_decode_xer,
	SEQUENCE_encode_xer,
	SEQUENCE_decode_uper,
	SEQUENCE_encode_uper,
	0,	/* Use generic outmost tag fetcher */
	asn_DEF_FANSPositionReport_tags_1,
	sizeof(asn_DEF_FANSPositionReport_tags_1)
		/sizeof(asn_DEF_FANSPositionReport_tags_1[0]), /* 1 */
	asn_DEF_FANSPositionReport_tags_1,	/* Same as above */
	sizeof(asn_DEF_FANSPositionReport_tags_1)
		/sizeof(asn_DEF_FANSPositionReport_tags_1[0]), /* 1 */
	0,	/* No PER visible constraints */
	asn_MBR_FANSPositionReport_1,
	22,	/* Elements count */
	&asn_SPC_FANSPositionReport_specs_1	/* Additional specs */
};

