#pragma once

#define SNPP_SCID 157
#define JPSS1_SCID 159
#define JPSS2_SCID 177
#define JPSS3_SCID 434
#define JPSS4_SCID 435

#define SNPP_NORAD 37849
#define JPSS1_NORAD 43013
#define JPSS2_NORAD 54234
#define JPSS3_NORAD 0 // Not launched yet
#define JPSS4_NORAD 0 // Not launched yet
