#pragma once

#define METOP_A_SCID 12
#define METOP_B_SCID 11
#define METOP_C_SCID 13

#define METOP_A_NORAD 29499
#define METOP_B_NORAD 38771
#define METOP_C_NORAD 43689

#define METOP_ORBIT_HEIGHT 827

#define METOP_AVHRR_SWATH 2900
#define METOP_AVHRR_RES 1.1

#define METOP_MHS_SWATH 2180
#define METOP_MHS_RES 16