/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSPositionReport_H_
#define	_FANSPositionReport_H_


#include "asn_application.h"

/* Including external dependencies */
#include "FANSPositionCurrent.h"
#include "FANSTimeAtPositionCurrent.h"
#include "FANSAltitude.h"
#include "FANSTurbulence.h"
#include "FANSIcing.h"
#include "FANSSpeedGround.h"
#include "FANSSupplementaryInformation.h"
#include "constr_SEQUENCE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Forward declarations */
struct FANSPosition;
struct FANSTime;
struct FANSRemainingFuel;
struct FANSTemperature;
struct FANSWinds;
struct FANSSpeed;
struct FANSVerticalChange;
struct FANSDegrees;
struct FANSDistance;
struct FANSAltitude;

/* FANSPositionReport */
typedef struct FANSPositionReport {
	FANSPositionCurrent_t	 positioncurrent;
	FANSTimeAtPositionCurrent_t	 timeatpositioncurrent;
	FANSAltitude_t	 altitude;
	struct FANSPosition	*fixnext	/* OPTIONAL */;
	struct FANSTime	*timeetaatfixnext	/* OPTIONAL */;
	struct FANSPosition	*fixnextplusone	/* OPTIONAL */;
	struct FANSTime	*timeetadestination	/* OPTIONAL */;
	struct FANSRemainingFuel	*remainingFuel	/* OPTIONAL */;
	struct FANSTemperature	*temperature	/* OPTIONAL */;
	struct FANSWinds	*winds	/* OPTIONAL */;
	FANSTurbulence_t	*turbulence	/* OPTIONAL */;
	FANSIcing_t	*icing	/* OPTIONAL */;
	struct FANSSpeed	*speed	/* OPTIONAL */;
	FANSSpeedGround_t	*speedground	/* OPTIONAL */;
	struct FANSVerticalChange	*verticalChange	/* OPTIONAL */;
	struct FANSDegrees	*trackAngle	/* OPTIONAL */;
	struct FANSDegrees	*trueHeading	/* OPTIONAL */;
	struct FANSDistance	*distance	/* OPTIONAL */;
	FANSSupplementaryInformation_t	*supplementaryInformation	/* OPTIONAL */;
	struct FANSPosition	*reportedWaypointPosition	/* OPTIONAL */;
	struct FANSTime	*reportedWaypointTime	/* OPTIONAL */;
	struct FANSAltitude	*reportedWaypointAltitude	/* OPTIONAL */;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} FANSPositionReport_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSPositionReport;

#ifdef __cplusplus
}
#endif

/* Referred external types */
#include "FANSFixNext.h"
#include "FANSTimeEtaAtFixNext.h"
#include "FANSFixNextPlusOne.h"
#include "FANSTimeEtaDestination.h"
#include "FANSRemainingFuel.h"
#include "FANSTemperature.h"
#include "FANSWinds.h"
#include "FANSSpeed.h"
#include "FANSVerticalChange.h"
#include "FANSTrackAngle.h"
#include "FANSTrueheading.h"
#include "FANSDistance.h"
#include "FANSReportedWaypointPosition.h"
#include "FANSReportedWaypointTime.h"
#include "FANSReportedWaypointAltitude.h"

#endif	/* _FANSPositionReport_H_ */
#include "asn_internal.h"
