/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSRouteClearance_H_
#define	_FANSRouteClearance_H_


#include "asn_application.h"

/* Including external dependencies */
#include "FANSAirportDeparture.h"
#include "FANSAirportDestination.h"
#include "FANSAirwayIntercept.h"
#include "constr_SEQUENCE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Forward declarations */
struct FANSRunway;
struct FANSProcedureName;
struct FANSRouteInformationSequence;
struct FANSRouteInformationAdditional;

/* FANSRouteClearance */
typedef struct FANSRouteClearance {
	FANSAirportDeparture_t	*airportDeparture	/* OPTIONAL */;
	FANSAirportDestination_t	*airportDestination	/* OPTIONAL */;
	struct FANSRunway	*runwayDeparture	/* OPTIONAL */;
	struct FANSProcedureName	*procedureDeparture	/* OPTIONAL */;
	struct FANSRunway	*runwayArrival	/* OPTIONAL */;
	struct FANSProcedureName	*procedureApproach	/* OPTIONAL */;
	struct FANSProcedureName	*procedureArrival	/* OPTIONAL */;
	FANSAirwayIntercept_t	*airwayIntercept	/* OPTIONAL */;
	struct FANSRouteInformationSequence	*routeinformation_seqOf	/* OPTIONAL */;
	struct FANSRouteInformationAdditional	*routeInformationAdditional	/* OPTIONAL */;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} FANSRouteClearance_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSRouteClearance;

#ifdef __cplusplus
}
#endif

/* Referred external types */
#include "FANSRunwayDeparture.h"
#include "FANSProcedureDeparture.h"
#include "FANSRunwayArrival.h"
#include "FANSProcedureApproach.h"
#include "FANSProcedureArrival.h"
#include "FANSRouteInformationSequence.h"
#include "FANSRouteInformationAdditional.h"

#endif	/* _FANSRouteClearance_H_ */
#include "asn_internal.h"
