set(ASN_MODULE_HEADERS
	FANSATCUplinkMessage.h
	FANSATCUplinkMsgElementIdSequence.h
	FANSATCMessageHeader.h
	FANSMsgIdentificationNumber.h
	FANSMsgReferenceNumber.h
	FANSTimestamp.h
	FANSTimehours.h
	FANSTimeminutes.h
	FANSTimeSeconds.h
	FANSATCUplinkMsgElementId.h
	FANSATCDownlinkMessage.h
	FANSATCDownlinkMsgElementIdSequence.h
	FANSATCDownlinkMsgElementId.h
	FANSAirport.h
	FANSAltimeter.h
	FANSAltimeterEnglish.h
	FANSAltimeterMetric.h
	FANSAltitude.h
	FANSAltitudeQNH.h
	FANSAltitudeQNHMeters.h
	FANSAltitudeQFE.h
	FANSAltitudeQFEMeters.h
	FANSAltitudeGNSSFeet.h
	FANSAltitudeGNSSMeters.h
	FANSAltitudeFlightLevel.h
	FANSAltitudeFlightLevelMetric.h
	FANSAltitudePosition.h
	FANSAltitudeAltitude.h
	FANSAltitudeSpeed.h
	FANSAltitudeSpeedSpeed.h
	FANSAltitudeTime.h
	FANSATISCode.h
	FANSATWAltitude.h
	FANSATWAltitudeTolerance.h
	FANSBeaconCode.h
	FANSBeaconCodeOctalDigit.h
	FANSDegrees.h
	FANSDegreesMagnetic.h
	FANSDegreesTrue.h
	FANSDirection.h
	FANSDirectionDegrees.h
	FANSDistanceOffset.h
	FANSDistanceOffsetNm.h
	FANSDistanceOffsetKm.h
	FANSDistanceOffsetDirection.h
	FANSErrorInformation.h
	FANSFixName.h
	FANSFreeText.h
	FANSFrequency.h
	FANSFrequencyhf.h
	FANSFrequencyvhf.h
	FANSFrequencyuhf.h
	FANSFrequencysatchannel.h
	FANSHoldClearance.h
	FANSICAOUnitName.h
	FANSICAOFacilityIdentification.h
	FANSICAOfacilityDesignation.h
	FANSICAOFacilityName.h
	FANSICAOFacilityFunction.h
	FANSICAOUnitNameFrequency.h
	FANSICAOFacilityDesignationTp4Table.h
	FANSLatitudeLongitude.h
	FANSLatitudeLongitudeSequence.h
	FANSLatitude.h
	FANSLatitudeDegrees.h
	FANSMinutesLatLon.h
	FANSLatitudeDirection.h
	FANSLongitude.h
	FANSLongitudeDegrees.h
	FANSLongitudeDirection.h
	FANSLegType.h
	FANSLegDistance.h
	FANSLegDistanceEnglish.h
	FANSLegDistanceMetric.h
	FANSLegTime.h
	FANSPlaceBearingDistance.h
	FANSDistance.h
	FANSDistanceNm.h
	FANSDistanceKm.h
	FANSPosition.h
	FANSNavaid.h
	FANSPositionAltitude.h
	FANSPositionAltitudeAltitude.h
	FANSPositionAltitudeSpeed.h
	FANSPositionDegrees.h
	FANSPositionDistanceOffsetDirection.h
	FANSPositionICAOUnitNameFrequency.h
	FANSPositionPosition.h
	FANSPositionProcedureName.h
	FANSPositionReport.h
	FANSPositionCurrent.h
	FANSTimeAtPositionCurrent.h
	FANSFixNext.h
	FANSTimeEtaAtFixNext.h
	FANSFixNextPlusOne.h
	FANSTimeEtaDestination.h
	FANSTemperature.h
	FANSTemperatureC.h
	FANSTemperatureF.h
	FANSWinds.h
	FANSWindDirection.h
	FANSWindSpeed.h
	FANSWindSpeedEnglish.h
	FANSWindSpeedMetric.h
	FANSTurbulence.h
	FANSIcing.h
	FANSVerticalChange.h
	FANSVerticalDirection.h
	FANSTrackAngle.h
	FANSTrueheading.h
	FANSSupplementaryInformation.h
	FANSReportedWaypointPosition.h
	FANSReportedWaypointTime.h
	FANSReportedWaypointAltitude.h
	FANSPositionRouteClearance.h
	FANSPositionSpeed.h
	FANSPositionSpeedSpeed.h
	FANSPositionTime.h
	FANSPositionTimeAltitude.h
	FANSPositionTimeTime.h
	FANSPredepartureClearance.h
	FANSAircraftFlightIdentification.h
	FANSAircraftType.h
	FANSAircraftEquipmentCode.h
	FANSCOMNAVApproachEquipmentAvailable.h
	FANSCOMNAVEquipmentStatus.h
	FANSCOMNAVEquipmentStatusSequence.h
	FANSSSREquipmentAvailable.h
	FANSTimeDepartureEdct.h
	FANSAltitudeRestriction.h
	FANSFrequencyDeparture.h
	FANSPDCrevision.h
	FANSProcedureName.h
	FANSProcedureType.h
	FANSProcedure.h
	FANSProcedureTransition.h
	FANSRemainingFuelRemainingSouls.h
	FANSRemainingFuel.h
	FANSRemainingSouls.h
	FANSRouteClearance.h
	FANSAirportDeparture.h
	FANSAirportDestination.h
	FANSRunwayDeparture.h
	FANSRunway.h
	FANSRunwayDirection.h
	FANSRunwayConfiguration.h
	FANSProcedureDeparture.h
	FANSRunwayArrival.h
	FANSProcedureApproach.h
	FANSProcedureArrival.h
	FANSAirwayIntercept.h
	FANSRouteInformation.h
	FANSRouteInformationSequence.h
	FANSPublishedIdentifier.h
	FANSPlaceBearing.h
	FANSPlaceBearingPlaceBearing.h
	FANSAirwayIdentifier.h
	FANSTrackDetail.h
	FANSTrackName.h
	FANSATWAlongTrackWaypointSequence.h
	FANSInterceptCourseFromSequence.h
	FANSHoldatwaypointSequence.h
	FANSWaypointSpeedAltitudesequence.h
	FANSRTARequiredTimeArrivalSequence.h
	FANSRouteInformationAdditional.h
	FANSATWAltitudeSequence.h
	FANSATWAlongTrackWaypoint.h
	FANSATWDistance.h
	FANSATWDistanceTolerance.h
	FANSReportingPoints.h
	FANSLatLonReportingPoints.h
	FANSLatitudeReportingPoints.h
	FANSLongitudeReportingPoints.h
	FANSDegreeIncrement.h
	FANSInterceptCourseFrom.h
	FANSInterceptCourseFromSelection.h
	FANSHoldatwaypointSpeedLow.h
	FANSHoldatwaypointSpeedHigh.h
	FANSEFCtime.h
	FANSHoldatwaypoint.h
	FANSWaypointSpeedAltitude.h
	FANSRTARequiredTimeArrival.h
	FANSRTATime.h
	FANSTimeTolerance.h
	FANSRTATolerance.h
	FANSSpeed.h
	FANSSpeedIndicated.h
	FANSSpeedIndicatedMetric.h
	FANSSpeedTrue.h
	FANSSpeedTrueMetric.h
	FANSSpeedGround.h
	FANSSpeedGroundMetric.h
	FANSSpeedMach.h
	FANSSpeedMachLarge.h
	FANSSpeedSpeed.h
	FANSTime.h
	FANSTimeAltitude.h
	FANSTimeDistanceOffsetDirection.h
	FANSTimeDistanceToFromPosition.h
	FANSTimeTime.h
	FANSToFrom.h
	FANSTimeICAOunitnameFrequency.h
	FANSTimePosition.h
	FANSTimePositionAltitude.h
	FANSTimePositionAltitudeSpeed.h
	FANSTimeSpeed.h
	FANSTimeSpeedSpeed.h
	FANSToFromPosition.h
	FANSTp4table.h
	FANSVersionNumber.h
	FANSVerticalRate.h
	FANSVerticalRateEnglish.h
	FANSVerticalRateMetric.h
	BOOLEAN.h
	INTEGER.h
	NativeEnumerated.h
	IA5String.h
	NULL.h
	NativeInteger.h
	NumericString.h
	asn_SEQUENCE_OF.h
	asn_SET_OF.h
	constr_CHOICE.h
	constr_SEQUENCE.h
	constr_SEQUENCE_OF.h
	constr_SET_OF.h
	asn_application.h
	asn_system.h
	asn_codecs.h
	asn_internal.h
	OCTET_STRING.h
	BIT_STRING.h
	asn_codecs_prim.h
	ber_tlv_length.h
	ber_tlv_tag.h
	ber_decoder.h
	der_encoder.h
	constr_TYPE.h
	constraints.h
	xer_support.h
	xer_decoder.h
	xer_encoder.h
	per_support.h
	per_decoder.h
	per_encoder.h
	per_opentype.h
)

set(ASN_MODULE_SOURCES
	FANSATCUplinkMessage.c
	FANSATCUplinkMsgElementIdSequence.c
	FANSATCMessageHeader.c
	FANSMsgIdentificationNumber.c
	FANSMsgReferenceNumber.c
	FANSTimestamp.c
	FANSTimehours.c
	FANSTimeminutes.c
	FANSTimeSeconds.c
	FANSATCUplinkMsgElementId.c
	FANSATCDownlinkMessage.c
	FANSATCDownlinkMsgElementIdSequence.c
	FANSATCDownlinkMsgElementId.c
	FANSAirport.c
	FANSAltimeter.c
	FANSAltimeterEnglish.c
	FANSAltimeterMetric.c
	FANSAltitude.c
	FANSAltitudeQNH.c
	FANSAltitudeQNHMeters.c
	FANSAltitudeQFE.c
	FANSAltitudeQFEMeters.c
	FANSAltitudeGNSSFeet.c
	FANSAltitudeGNSSMeters.c
	FANSAltitudeFlightLevel.c
	FANSAltitudeFlightLevelMetric.c
	FANSAltitudePosition.c
	FANSAltitudeAltitude.c
	FANSAltitudeSpeed.c
	FANSAltitudeSpeedSpeed.c
	FANSAltitudeTime.c
	FANSATISCode.c
	FANSATWAltitude.c
	FANSATWAltitudeTolerance.c
	FANSBeaconCode.c
	FANSBeaconCodeOctalDigit.c
	FANSDegrees.c
	FANSDegreesMagnetic.c
	FANSDegreesTrue.c
	FANSDirection.c
	FANSDirectionDegrees.c
	FANSDistanceOffset.c
	FANSDistanceOffsetNm.c
	FANSDistanceOffsetKm.c
	FANSDistanceOffsetDirection.c
	FANSErrorInformation.c
	FANSFixName.c
	FANSFreeText.c
	FANSFrequency.c
	FANSFrequencyhf.c
	FANSFrequencyvhf.c
	FANSFrequencyuhf.c
	FANSFrequencysatchannel.c
	FANSHoldClearance.c
	FANSICAOUnitName.c
	FANSICAOFacilityIdentification.c
	FANSICAOfacilityDesignation.c
	FANSICAOFacilityName.c
	FANSICAOFacilityFunction.c
	FANSICAOUnitNameFrequency.c
	FANSICAOFacilityDesignationTp4Table.c
	FANSLatitudeLongitude.c
	FANSLatitudeLongitudeSequence.c
	FANSLatitude.c
	FANSLatitudeDegrees.c
	FANSMinutesLatLon.c
	FANSLatitudeDirection.c
	FANSLongitude.c
	FANSLongitudeDegrees.c
	FANSLongitudeDirection.c
	FANSLegType.c
	FANSLegDistance.c
	FANSLegDistanceEnglish.c
	FANSLegDistanceMetric.c
	FANSLegTime.c
	FANSPlaceBearingDistance.c
	FANSDistance.c
	FANSDistanceNm.c
	FANSDistanceKm.c
	FANSPosition.c
	FANSNavaid.c
	FANSPositionAltitude.c
	FANSPositionAltitudeAltitude.c
	FANSPositionAltitudeSpeed.c
	FANSPositionDegrees.c
	FANSPositionDistanceOffsetDirection.c
	FANSPositionICAOUnitNameFrequency.c
	FANSPositionPosition.c
	FANSPositionProcedureName.c
	FANSPositionReport.c
	FANSPositionCurrent.c
	FANSTimeAtPositionCurrent.c
	FANSFixNext.c
	FANSTimeEtaAtFixNext.c
	FANSFixNextPlusOne.c
	FANSTimeEtaDestination.c
	FANSTemperature.c
	FANSTemperatureC.c
	FANSTemperatureF.c
	FANSWinds.c
	FANSWindDirection.c
	FANSWindSpeed.c
	FANSWindSpeedEnglish.c
	FANSWindSpeedMetric.c
	FANSTurbulence.c
	FANSIcing.c
	FANSVerticalChange.c
	FANSVerticalDirection.c
	FANSTrackAngle.c
	FANSTrueheading.c
	FANSSupplementaryInformation.c
	FANSReportedWaypointPosition.c
	FANSReportedWaypointTime.c
	FANSReportedWaypointAltitude.c
	FANSPositionRouteClearance.c
	FANSPositionSpeed.c
	FANSPositionSpeedSpeed.c
	FANSPositionTime.c
	FANSPositionTimeAltitude.c
	FANSPositionTimeTime.c
	FANSPredepartureClearance.c
	FANSAircraftFlightIdentification.c
	FANSAircraftType.c
	FANSAircraftEquipmentCode.c
	FANSCOMNAVApproachEquipmentAvailable.c
	FANSCOMNAVEquipmentStatus.c
	FANSCOMNAVEquipmentStatusSequence.c
	FANSSSREquipmentAvailable.c
	FANSTimeDepartureEdct.c
	FANSAltitudeRestriction.c
	FANSFrequencyDeparture.c
	FANSPDCrevision.c
	FANSProcedureName.c
	FANSProcedureType.c
	FANSProcedure.c
	FANSProcedureTransition.c
	FANSRemainingFuelRemainingSouls.c
	FANSRemainingFuel.c
	FANSRemainingSouls.c
	FANSRouteClearance.c
	FANSAirportDeparture.c
	FANSAirportDestination.c
	FANSRunwayDeparture.c
	FANSRunway.c
	FANSRunwayDirection.c
	FANSRunwayConfiguration.c
	FANSProcedureDeparture.c
	FANSRunwayArrival.c
	FANSProcedureApproach.c
	FANSProcedureArrival.c
	FANSAirwayIntercept.c
	FANSRouteInformation.c
	FANSRouteInformationSequence.c
	FANSPublishedIdentifier.c
	FANSPlaceBearing.c
	FANSPlaceBearingPlaceBearing.c
	FANSAirwayIdentifier.c
	FANSTrackDetail.c
	FANSTrackName.c
	FANSATWAlongTrackWaypointSequence.c
	FANSInterceptCourseFromSequence.c
	FANSHoldatwaypointSequence.c
	FANSWaypointSpeedAltitudesequence.c
	FANSRTARequiredTimeArrivalSequence.c
	FANSRouteInformationAdditional.c
	FANSATWAltitudeSequence.c
	FANSATWAlongTrackWaypoint.c
	FANSATWDistance.c
	FANSATWDistanceTolerance.c
	FANSReportingPoints.c
	FANSLatLonReportingPoints.c
	FANSLatitudeReportingPoints.c
	FANSLongitudeReportingPoints.c
	FANSDegreeIncrement.c
	FANSInterceptCourseFrom.c
	FANSInterceptCourseFromSelection.c
	FANSHoldatwaypointSpeedLow.c
	FANSHoldatwaypointSpeedHigh.c
	FANSEFCtime.c
	FANSHoldatwaypoint.c
	FANSWaypointSpeedAltitude.c
	FANSRTARequiredTimeArrival.c
	FANSRTATime.c
	FANSTimeTolerance.c
	FANSRTATolerance.c
	FANSSpeed.c
	FANSSpeedIndicated.c
	FANSSpeedIndicatedMetric.c
	FANSSpeedTrue.c
	FANSSpeedTrueMetric.c
	FANSSpeedGround.c
	FANSSpeedGroundMetric.c
	FANSSpeedMach.c
	FANSSpeedMachLarge.c
	FANSSpeedSpeed.c
	FANSTime.c
	FANSTimeAltitude.c
	FANSTimeDistanceOffsetDirection.c
	FANSTimeDistanceToFromPosition.c
	FANSTimeTime.c
	FANSToFrom.c
	FANSTimeICAOunitnameFrequency.c
	FANSTimePosition.c
	FANSTimePositionAltitude.c
	FANSTimePositionAltitudeSpeed.c
	FANSTimeSpeed.c
	FANSTimeSpeedSpeed.c
	FANSToFromPosition.c
	FANSTp4table.c
	FANSVersionNumber.c
	FANSVerticalRate.c
	FANSVerticalRateEnglish.c
	FANSVerticalRateMetric.c
	BOOLEAN.c
	IA5String.c
	INTEGER.c
	NULL.c
	NativeEnumerated.c
	NativeInteger.c
	NumericString.c
	asn_SEQUENCE_OF.c
	asn_SET_OF.c
	constr_CHOICE.c
	constr_SEQUENCE.c
	constr_SEQUENCE_OF.c
	constr_SET_OF.c
	OCTET_STRING.c
	BIT_STRING.c
	asn_codecs_prim.c
	ber_tlv_length.c
	ber_tlv_tag.c
	ber_decoder.c
	der_encoder.c
	constr_TYPE.c
	constraints.c
	xer_support.c
	xer_decoder.c
	xer_encoder.c
	per_support.c
	per_decoder.c
	per_encoder.c
	per_opentype.c
)

