/*
 *  This file is a part of libacars
 *
 *  Copyright (c) 2018-2021 Tomasz Lemiech <szpajder@gmail.com>
 */

#ifndef LA_VERSION_H
#define LA_VERSION_H 1

#ifdef __cplusplus
extern "C" {
#endif

extern char const * const LA_VERSION;

#ifdef __cplusplus
}
#endif

#endif // !LA_VERSION_H
