/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#include "FANSATCUplinkMsgElementId.h"

static asn_per_constraints_t asn_PER_type_FANSATCUplinkMsgElementId_constr_1 GCC_NOTUSED = {
	{ APC_CONSTRAINED,	 8,  8,  0,  182 }	/* (0..182) */,
	{ APC_UNCONSTRAINED,	-1, -1,  0,  0 },
	0, 0	/* No PER value map */
};
static asn_TYPE_member_t asn_MBR_FANSATCUplinkMsgElementId_1[] = {
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM0NULL),
		(ASN_TAG_CLASS_CONTEXT | (0 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM0NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM1NULL),
		(ASN_TAG_CLASS_CONTEXT | (1 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM1NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM2NULL),
		(ASN_TAG_CLASS_CONTEXT | (2 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM2NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM3NULL),
		(ASN_TAG_CLASS_CONTEXT | (3 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM3NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM4NULL),
		(ASN_TAG_CLASS_CONTEXT | (4 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM4NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM5NULL),
		(ASN_TAG_CLASS_CONTEXT | (5 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM5NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM6Altitude),
		(ASN_TAG_CLASS_CONTEXT | (6 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM6Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM7Time),
		(ASN_TAG_CLASS_CONTEXT | (7 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM7Time"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM8Position),
		(ASN_TAG_CLASS_CONTEXT | (8 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM8Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM9Time),
		(ASN_TAG_CLASS_CONTEXT | (9 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM9Time"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM10Position),
		(ASN_TAG_CLASS_CONTEXT | (10 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM10Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM11Time),
		(ASN_TAG_CLASS_CONTEXT | (11 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM11Time"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM12Position),
		(ASN_TAG_CLASS_CONTEXT | (12 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM12Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM13TimeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (13 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM13TimeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM14PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (14 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM14PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM15TimeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (15 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM15TimeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM16PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (16 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM16PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM17TimeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (17 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM17TimeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM18PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (18 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM18PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM19Altitude),
		(ASN_TAG_CLASS_CONTEXT | (19 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM19Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM20Altitude),
		(ASN_TAG_CLASS_CONTEXT | (20 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM20Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM21TimeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (21 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM21TimeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM22PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (22 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM22PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM23Altitude),
		(ASN_TAG_CLASS_CONTEXT | (23 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM23Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM24TimeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (24 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM24TimeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM25PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (25 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM25PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM26AltitudeTime),
		(ASN_TAG_CLASS_CONTEXT | (26 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM26AltitudeTime"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM27AltitudePosition),
		(ASN_TAG_CLASS_CONTEXT | (27 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudePosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM27AltitudePosition"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM28AltitudeTime),
		(ASN_TAG_CLASS_CONTEXT | (28 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM28AltitudeTime"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM29AltitudePosition),
		(ASN_TAG_CLASS_CONTEXT | (29 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudePosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM29AltitudePosition"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM30AltitudeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (30 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM30AltitudeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM31AltitudeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (31 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM31AltitudeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM32AltitudeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (32 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM32AltitudeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM33Altitude),
		(ASN_TAG_CLASS_CONTEXT | (33 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM33Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM34Altitude),
		(ASN_TAG_CLASS_CONTEXT | (34 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM34Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM35Altitude),
		(ASN_TAG_CLASS_CONTEXT | (35 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM35Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM36Altitude),
		(ASN_TAG_CLASS_CONTEXT | (36 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM36Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM37Altitude),
		(ASN_TAG_CLASS_CONTEXT | (37 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM37Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM38Altitude),
		(ASN_TAG_CLASS_CONTEXT | (38 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM38Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM39Altitude),
		(ASN_TAG_CLASS_CONTEXT | (39 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM39Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM40Altitude),
		(ASN_TAG_CLASS_CONTEXT | (40 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM40Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM41Altitude),
		(ASN_TAG_CLASS_CONTEXT | (41 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM41Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM42PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (42 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM42PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM43PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (43 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM43PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM44PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (44 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM44PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM45PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (45 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM45PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM46PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (46 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM46PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM47PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (47 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM47PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM48PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (48 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM48PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM49PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (49 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM49PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM50PositionAltitudeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (50 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitudeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM50PositionAltitudeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM51PositionTime),
		(ASN_TAG_CLASS_CONTEXT | (51 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM51PositionTime"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM52PositionTime),
		(ASN_TAG_CLASS_CONTEXT | (52 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM52PositionTime"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM53PositionTime),
		(ASN_TAG_CLASS_CONTEXT | (53 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM53PositionTime"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM54PositionTimeTime),
		(ASN_TAG_CLASS_CONTEXT | (54 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionTimeTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM54PositionTimeTime"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM55PositionSpeed),
		(ASN_TAG_CLASS_CONTEXT | (55 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM55PositionSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM56PositionSpeed),
		(ASN_TAG_CLASS_CONTEXT | (56 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM56PositionSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM57PositionSpeed),
		(ASN_TAG_CLASS_CONTEXT | (57 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM57PositionSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM58PositionTimeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (58 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionTimeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM58PositionTimeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM59PositionTimeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (59 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionTimeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM59PositionTimeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM60PositionTimeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (60 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionTimeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM60PositionTimeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM61PositionAltitudeSpeed),
		(ASN_TAG_CLASS_CONTEXT | (61 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitudeSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM61PositionAltitudeSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM62TimePositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (62 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimePositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM62TimePositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM63TimePositionAltitudeSpeed),
		(ASN_TAG_CLASS_CONTEXT | (63 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimePositionAltitudeSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM63TimePositionAltitudeSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM64DistanceOffsetDirection),
		(ASN_TAG_CLASS_CONTEXT | (64 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSDistanceOffsetDirection,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM64DistanceOffsetDirection"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM65PositionDistanceOffsetDirection),
		(ASN_TAG_CLASS_CONTEXT | (65 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionDistanceOffsetDirection,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM65PositionDistanceOffsetDirection"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM66TimeDistanceOffsetDirection),
		(ASN_TAG_CLASS_CONTEXT | (66 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeDistanceOffsetDirection,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM66TimeDistanceOffsetDirection"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM67NULL),
		(ASN_TAG_CLASS_CONTEXT | (67 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM67NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM68Position),
		(ASN_TAG_CLASS_CONTEXT | (68 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM68Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM69Time),
		(ASN_TAG_CLASS_CONTEXT | (69 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM69Time"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM70Position),
		(ASN_TAG_CLASS_CONTEXT | (70 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM70Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM71Time),
		(ASN_TAG_CLASS_CONTEXT | (71 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM71Time"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM72NULL),
		(ASN_TAG_CLASS_CONTEXT | (72 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM72NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM73Predepartureclearance),
		(ASN_TAG_CLASS_CONTEXT | (73 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPredepartureClearance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM73Predepartureclearance"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM74Position),
		(ASN_TAG_CLASS_CONTEXT | (74 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM74Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM75Position),
		(ASN_TAG_CLASS_CONTEXT | (75 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM75Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM76TimePosition),
		(ASN_TAG_CLASS_CONTEXT | (76 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimePosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM76TimePosition"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM77PositionPosition),
		(ASN_TAG_CLASS_CONTEXT | (77 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM77PositionPosition"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM78AltitudePosition),
		(ASN_TAG_CLASS_CONTEXT | (78 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudePosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM78AltitudePosition"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM79PositionRouteClearance),
		(ASN_TAG_CLASS_CONTEXT | (79 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionRouteClearance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM79PositionRouteClearance"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM80RouteClearance),
		(ASN_TAG_CLASS_CONTEXT | (80 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSRouteClearance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM80RouteClearance"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM81ProcedureName),
		(ASN_TAG_CLASS_CONTEXT | (81 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSProcedureName,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM81ProcedureName"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM82DistanceOffsetDirection),
		(ASN_TAG_CLASS_CONTEXT | (82 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSDistanceOffsetDirection,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM82DistanceOffsetDirection"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM83PositionRouteClearance),
		(ASN_TAG_CLASS_CONTEXT | (83 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionRouteClearance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM83PositionRouteClearance"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM84PositionProcedureName),
		(ASN_TAG_CLASS_CONTEXT | (84 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionProcedureName,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM84PositionProcedureName"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM85RouteClearance),
		(ASN_TAG_CLASS_CONTEXT | (85 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSRouteClearance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM85RouteClearance"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM86PositionRouteClearance),
		(ASN_TAG_CLASS_CONTEXT | (86 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionRouteClearance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM86PositionRouteClearance"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM87Position),
		(ASN_TAG_CLASS_CONTEXT | (87 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM87Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM88PositionPosition),
		(ASN_TAG_CLASS_CONTEXT | (88 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM88PositionPosition"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM89TimePosition),
		(ASN_TAG_CLASS_CONTEXT | (89 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimePosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM89TimePosition"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM90AltitudePosition),
		(ASN_TAG_CLASS_CONTEXT | (90 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudePosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM90AltitudePosition"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM91HoldClearance),
		(ASN_TAG_CLASS_CONTEXT | (91 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSHoldClearance,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM91HoldClearance"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM92PositionAltitude),
		(ASN_TAG_CLASS_CONTEXT | (92 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM92PositionAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM93Time),
		(ASN_TAG_CLASS_CONTEXT | (93 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM93Time"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM94DirectionDegrees),
		(ASN_TAG_CLASS_CONTEXT | (94 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSDirectionDegrees,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM94DirectionDegrees"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM95DirectionDegrees),
		(ASN_TAG_CLASS_CONTEXT | (95 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSDirectionDegrees,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM95DirectionDegrees"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM96NULL),
		(ASN_TAG_CLASS_CONTEXT | (96 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM96NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM97PositionDegrees),
		(ASN_TAG_CLASS_CONTEXT | (97 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionDegrees,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM97PositionDegrees"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM98DirectionDegrees),
		(ASN_TAG_CLASS_CONTEXT | (98 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSDirectionDegrees,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM98DirectionDegrees"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM99ProcedureName),
		(ASN_TAG_CLASS_CONTEXT | (99 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSProcedureName,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM99ProcedureName"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM100TimeSpeed),
		(ASN_TAG_CLASS_CONTEXT | (100 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM100TimeSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM101PositionSpeed),
		(ASN_TAG_CLASS_CONTEXT | (101 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM101PositionSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM102AltitudeSpeed),
		(ASN_TAG_CLASS_CONTEXT | (102 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM102AltitudeSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM103TimeSpeedSpeed),
		(ASN_TAG_CLASS_CONTEXT | (103 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeSpeedSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM103TimeSpeedSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM104PositionSpeedSpeed),
		(ASN_TAG_CLASS_CONTEXT | (104 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionSpeedSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM104PositionSpeedSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM105AltitudeSpeedSpeed),
		(ASN_TAG_CLASS_CONTEXT | (105 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeSpeedSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM105AltitudeSpeedSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM106Speed),
		(ASN_TAG_CLASS_CONTEXT | (106 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM106Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM107NULL),
		(ASN_TAG_CLASS_CONTEXT | (107 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM107NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM108Speed),
		(ASN_TAG_CLASS_CONTEXT | (108 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM108Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM109Speed),
		(ASN_TAG_CLASS_CONTEXT | (109 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM109Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM110SpeedSpeed),
		(ASN_TAG_CLASS_CONTEXT | (110 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSSpeedSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM110SpeedSpeed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM111Speed),
		(ASN_TAG_CLASS_CONTEXT | (111 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM111Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM112Speed),
		(ASN_TAG_CLASS_CONTEXT | (112 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM112Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM113Speed),
		(ASN_TAG_CLASS_CONTEXT | (113 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM113Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM114Speed),
		(ASN_TAG_CLASS_CONTEXT | (114 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM114Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM115Speed),
		(ASN_TAG_CLASS_CONTEXT | (115 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM115Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM116NULL),
		(ASN_TAG_CLASS_CONTEXT | (116 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM116NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM117ICAOunitnameFrequency),
		(ASN_TAG_CLASS_CONTEXT | (117 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSICAOUnitNameFrequency,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM117ICAOunitnameFrequency"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM118PositionICAOunitnameFrequency),
		(ASN_TAG_CLASS_CONTEXT | (118 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionICAOUnitNameFrequency,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM118PositionICAOunitnameFrequency"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM119TimeICAOunitnameFrequency),
		(ASN_TAG_CLASS_CONTEXT | (119 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeICAOunitnameFrequency,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM119TimeICAOunitnameFrequency"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM120ICAOunitnameFrequency),
		(ASN_TAG_CLASS_CONTEXT | (120 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSICAOUnitNameFrequency,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM120ICAOunitnameFrequency"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM121PositionICAOunitnameFrequency),
		(ASN_TAG_CLASS_CONTEXT | (121 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSPositionICAOUnitNameFrequency,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM121PositionICAOunitnameFrequency"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM122TimeICAOunitnameFrequency),
		(ASN_TAG_CLASS_CONTEXT | (122 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSTimeICAOunitnameFrequency,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM122TimeICAOunitnameFrequency"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM123BeaconCode),
		(ASN_TAG_CLASS_CONTEXT | (123 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSBeaconCode,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM123BeaconCode"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM124NULL),
		(ASN_TAG_CLASS_CONTEXT | (124 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM124NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM125NULL),
		(ASN_TAG_CLASS_CONTEXT | (125 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM125NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM126NULL),
		(ASN_TAG_CLASS_CONTEXT | (126 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM126NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM127NULL),
		(ASN_TAG_CLASS_CONTEXT | (127 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM127NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM128Altitude),
		(ASN_TAG_CLASS_CONTEXT | (128 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM128Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM129Altitude),
		(ASN_TAG_CLASS_CONTEXT | (129 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM129Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM130Position),
		(ASN_TAG_CLASS_CONTEXT | (130 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM130Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM131NULL),
		(ASN_TAG_CLASS_CONTEXT | (131 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM131NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM132NULL),
		(ASN_TAG_CLASS_CONTEXT | (132 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM132NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM133NULL),
		(ASN_TAG_CLASS_CONTEXT | (133 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM133NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM134NULL),
		(ASN_TAG_CLASS_CONTEXT | (134 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM134NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM135NULL),
		(ASN_TAG_CLASS_CONTEXT | (135 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM135NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM136NULL),
		(ASN_TAG_CLASS_CONTEXT | (136 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM136NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM137NULL),
		(ASN_TAG_CLASS_CONTEXT | (137 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM137NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM138NULL),
		(ASN_TAG_CLASS_CONTEXT | (138 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM138NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM139NULL),
		(ASN_TAG_CLASS_CONTEXT | (139 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM139NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM140NULL),
		(ASN_TAG_CLASS_CONTEXT | (140 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM140NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM141NULL),
		(ASN_TAG_CLASS_CONTEXT | (141 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM141NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM142NULL),
		(ASN_TAG_CLASS_CONTEXT | (142 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM142NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM143NULL),
		(ASN_TAG_CLASS_CONTEXT | (143 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM143NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM144NULL),
		(ASN_TAG_CLASS_CONTEXT | (144 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM144NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM145NULL),
		(ASN_TAG_CLASS_CONTEXT | (145 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM145NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM146NULL),
		(ASN_TAG_CLASS_CONTEXT | (146 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM146NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM147NULL),
		(ASN_TAG_CLASS_CONTEXT | (147 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM147NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM148Altitude),
		(ASN_TAG_CLASS_CONTEXT | (148 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM148Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM149AltitudePosition),
		(ASN_TAG_CLASS_CONTEXT | (149 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudePosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM149AltitudePosition"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM150AltitudeTime),
		(ASN_TAG_CLASS_CONTEXT | (150 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeTime,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM150AltitudeTime"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM151Speed),
		(ASN_TAG_CLASS_CONTEXT | (151 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSSpeed,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM151Speed"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM152DistanceOffsetDirection),
		(ASN_TAG_CLASS_CONTEXT | (152 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSDistanceOffsetDirection,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM152DistanceOffsetDirection"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM153Altimeter),
		(ASN_TAG_CLASS_CONTEXT | (153 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltimeter,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM153Altimeter"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM154NULL),
		(ASN_TAG_CLASS_CONTEXT | (154 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM154NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM155Position),
		(ASN_TAG_CLASS_CONTEXT | (155 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM155Position"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM156NULL),
		(ASN_TAG_CLASS_CONTEXT | (156 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM156NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM157Frequency),
		(ASN_TAG_CLASS_CONTEXT | (157 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSFrequency,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM157Frequency"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM158ATISCode),
		(ASN_TAG_CLASS_CONTEXT | (158 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSATISCode,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM158ATISCode"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM159ErrorInformation),
		(ASN_TAG_CLASS_CONTEXT | (159 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSErrorInformation,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM159ErrorInformation"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM160ICAOfacilitydesignation),
		(ASN_TAG_CLASS_CONTEXT | (160 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSICAOfacilityDesignation,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM160ICAOfacilitydesignation"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM161NULL),
		(ASN_TAG_CLASS_CONTEXT | (161 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM161NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM162NULL),
		(ASN_TAG_CLASS_CONTEXT | (162 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM162NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM163ICAOfacilitydesignationTp4table),
		(ASN_TAG_CLASS_CONTEXT | (163 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSICAOFacilityDesignationTp4Table,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM163ICAOfacilitydesignationTp4table"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM164NULL),
		(ASN_TAG_CLASS_CONTEXT | (164 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM164NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM165NULL),
		(ASN_TAG_CLASS_CONTEXT | (165 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM165NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM166NULL),
		(ASN_TAG_CLASS_CONTEXT | (166 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM166NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM167NULL),
		(ASN_TAG_CLASS_CONTEXT | (167 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM167NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM168NULL),
		(ASN_TAG_CLASS_CONTEXT | (168 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM168NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM169FreeText),
		(ASN_TAG_CLASS_CONTEXT | (169 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSFreeText,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM169FreeText"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM170FreeText),
		(ASN_TAG_CLASS_CONTEXT | (170 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSFreeText,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM170FreeText"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM171VerticalRate),
		(ASN_TAG_CLASS_CONTEXT | (171 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSVerticalRate,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM171VerticalRate"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM172VerticalRate),
		(ASN_TAG_CLASS_CONTEXT | (172 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSVerticalRate,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM172VerticalRate"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM173VerticalRate),
		(ASN_TAG_CLASS_CONTEXT | (173 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSVerticalRate,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM173VerticalRate"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM174VerticalRate),
		(ASN_TAG_CLASS_CONTEXT | (174 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSVerticalRate,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM174VerticalRate"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM175Altitude),
		(ASN_TAG_CLASS_CONTEXT | (175 << 2)),
		+1,	/* EXPLICIT tag at current level */
		&asn_DEF_FANSAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM175Altitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM176NULL),
		(ASN_TAG_CLASS_CONTEXT | (176 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM176NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM177NULL),
		(ASN_TAG_CLASS_CONTEXT | (177 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM177NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM178NULL),
		(ASN_TAG_CLASS_CONTEXT | (178 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM178NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM179NULL),
		(ASN_TAG_CLASS_CONTEXT | (179 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM179NULL"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM180AltitudeAltitude),
		(ASN_TAG_CLASS_CONTEXT | (180 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAltitudeAltitude,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM180AltitudeAltitude"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM181ToFromPosition),
		(ASN_TAG_CLASS_CONTEXT | (181 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSToFromPosition,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM181ToFromPosition"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct FANSATCUplinkMsgElementId, choice.uM182NULL),
		(ASN_TAG_CLASS_CONTEXT | (182 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_NULL,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"uM182NULL"
		},
};
static const asn_TYPE_tag2member_t asn_MAP_FANSATCUplinkMsgElementId_tag2el_1[] = {
    { (ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0 }, /* uM0NULL */
    { (ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0 }, /* uM1NULL */
    { (ASN_TAG_CLASS_CONTEXT | (2 << 2)), 2, 0, 0 }, /* uM2NULL */
    { (ASN_TAG_CLASS_CONTEXT | (3 << 2)), 3, 0, 0 }, /* uM3NULL */
    { (ASN_TAG_CLASS_CONTEXT | (4 << 2)), 4, 0, 0 }, /* uM4NULL */
    { (ASN_TAG_CLASS_CONTEXT | (5 << 2)), 5, 0, 0 }, /* uM5NULL */
    { (ASN_TAG_CLASS_CONTEXT | (6 << 2)), 6, 0, 0 }, /* uM6Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (7 << 2)), 7, 0, 0 }, /* uM7Time */
    { (ASN_TAG_CLASS_CONTEXT | (8 << 2)), 8, 0, 0 }, /* uM8Position */
    { (ASN_TAG_CLASS_CONTEXT | (9 << 2)), 9, 0, 0 }, /* uM9Time */
    { (ASN_TAG_CLASS_CONTEXT | (10 << 2)), 10, 0, 0 }, /* uM10Position */
    { (ASN_TAG_CLASS_CONTEXT | (11 << 2)), 11, 0, 0 }, /* uM11Time */
    { (ASN_TAG_CLASS_CONTEXT | (12 << 2)), 12, 0, 0 }, /* uM12Position */
    { (ASN_TAG_CLASS_CONTEXT | (13 << 2)), 13, 0, 0 }, /* uM13TimeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (14 << 2)), 14, 0, 0 }, /* uM14PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (15 << 2)), 15, 0, 0 }, /* uM15TimeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (16 << 2)), 16, 0, 0 }, /* uM16PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (17 << 2)), 17, 0, 0 }, /* uM17TimeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (18 << 2)), 18, 0, 0 }, /* uM18PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (19 << 2)), 19, 0, 0 }, /* uM19Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (20 << 2)), 20, 0, 0 }, /* uM20Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (21 << 2)), 21, 0, 0 }, /* uM21TimeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (22 << 2)), 22, 0, 0 }, /* uM22PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (23 << 2)), 23, 0, 0 }, /* uM23Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (24 << 2)), 24, 0, 0 }, /* uM24TimeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (25 << 2)), 25, 0, 0 }, /* uM25PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (26 << 2)), 26, 0, 0 }, /* uM26AltitudeTime */
    { (ASN_TAG_CLASS_CONTEXT | (27 << 2)), 27, 0, 0 }, /* uM27AltitudePosition */
    { (ASN_TAG_CLASS_CONTEXT | (28 << 2)), 28, 0, 0 }, /* uM28AltitudeTime */
    { (ASN_TAG_CLASS_CONTEXT | (29 << 2)), 29, 0, 0 }, /* uM29AltitudePosition */
    { (ASN_TAG_CLASS_CONTEXT | (30 << 2)), 30, 0, 0 }, /* uM30AltitudeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (31 << 2)), 31, 0, 0 }, /* uM31AltitudeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (32 << 2)), 32, 0, 0 }, /* uM32AltitudeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (33 << 2)), 33, 0, 0 }, /* uM33Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (34 << 2)), 34, 0, 0 }, /* uM34Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (35 << 2)), 35, 0, 0 }, /* uM35Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (36 << 2)), 36, 0, 0 }, /* uM36Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (37 << 2)), 37, 0, 0 }, /* uM37Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (38 << 2)), 38, 0, 0 }, /* uM38Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (39 << 2)), 39, 0, 0 }, /* uM39Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (40 << 2)), 40, 0, 0 }, /* uM40Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (41 << 2)), 41, 0, 0 }, /* uM41Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (42 << 2)), 42, 0, 0 }, /* uM42PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (43 << 2)), 43, 0, 0 }, /* uM43PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (44 << 2)), 44, 0, 0 }, /* uM44PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (45 << 2)), 45, 0, 0 }, /* uM45PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (46 << 2)), 46, 0, 0 }, /* uM46PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (47 << 2)), 47, 0, 0 }, /* uM47PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (48 << 2)), 48, 0, 0 }, /* uM48PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (49 << 2)), 49, 0, 0 }, /* uM49PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (50 << 2)), 50, 0, 0 }, /* uM50PositionAltitudeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (51 << 2)), 51, 0, 0 }, /* uM51PositionTime */
    { (ASN_TAG_CLASS_CONTEXT | (52 << 2)), 52, 0, 0 }, /* uM52PositionTime */
    { (ASN_TAG_CLASS_CONTEXT | (53 << 2)), 53, 0, 0 }, /* uM53PositionTime */
    { (ASN_TAG_CLASS_CONTEXT | (54 << 2)), 54, 0, 0 }, /* uM54PositionTimeTime */
    { (ASN_TAG_CLASS_CONTEXT | (55 << 2)), 55, 0, 0 }, /* uM55PositionSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (56 << 2)), 56, 0, 0 }, /* uM56PositionSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (57 << 2)), 57, 0, 0 }, /* uM57PositionSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (58 << 2)), 58, 0, 0 }, /* uM58PositionTimeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (59 << 2)), 59, 0, 0 }, /* uM59PositionTimeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (60 << 2)), 60, 0, 0 }, /* uM60PositionTimeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (61 << 2)), 61, 0, 0 }, /* uM61PositionAltitudeSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (62 << 2)), 62, 0, 0 }, /* uM62TimePositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (63 << 2)), 63, 0, 0 }, /* uM63TimePositionAltitudeSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (64 << 2)), 64, 0, 0 }, /* uM64DistanceOffsetDirection */
    { (ASN_TAG_CLASS_CONTEXT | (65 << 2)), 65, 0, 0 }, /* uM65PositionDistanceOffsetDirection */
    { (ASN_TAG_CLASS_CONTEXT | (66 << 2)), 66, 0, 0 }, /* uM66TimeDistanceOffsetDirection */
    { (ASN_TAG_CLASS_CONTEXT | (67 << 2)), 67, 0, 0 }, /* uM67NULL */
    { (ASN_TAG_CLASS_CONTEXT | (68 << 2)), 68, 0, 0 }, /* uM68Position */
    { (ASN_TAG_CLASS_CONTEXT | (69 << 2)), 69, 0, 0 }, /* uM69Time */
    { (ASN_TAG_CLASS_CONTEXT | (70 << 2)), 70, 0, 0 }, /* uM70Position */
    { (ASN_TAG_CLASS_CONTEXT | (71 << 2)), 71, 0, 0 }, /* uM71Time */
    { (ASN_TAG_CLASS_CONTEXT | (72 << 2)), 72, 0, 0 }, /* uM72NULL */
    { (ASN_TAG_CLASS_CONTEXT | (73 << 2)), 73, 0, 0 }, /* uM73Predepartureclearance */
    { (ASN_TAG_CLASS_CONTEXT | (74 << 2)), 74, 0, 0 }, /* uM74Position */
    { (ASN_TAG_CLASS_CONTEXT | (75 << 2)), 75, 0, 0 }, /* uM75Position */
    { (ASN_TAG_CLASS_CONTEXT | (76 << 2)), 76, 0, 0 }, /* uM76TimePosition */
    { (ASN_TAG_CLASS_CONTEXT | (77 << 2)), 77, 0, 0 }, /* uM77PositionPosition */
    { (ASN_TAG_CLASS_CONTEXT | (78 << 2)), 78, 0, 0 }, /* uM78AltitudePosition */
    { (ASN_TAG_CLASS_CONTEXT | (79 << 2)), 79, 0, 0 }, /* uM79PositionRouteClearance */
    { (ASN_TAG_CLASS_CONTEXT | (80 << 2)), 80, 0, 0 }, /* uM80RouteClearance */
    { (ASN_TAG_CLASS_CONTEXT | (81 << 2)), 81, 0, 0 }, /* uM81ProcedureName */
    { (ASN_TAG_CLASS_CONTEXT | (82 << 2)), 82, 0, 0 }, /* uM82DistanceOffsetDirection */
    { (ASN_TAG_CLASS_CONTEXT | (83 << 2)), 83, 0, 0 }, /* uM83PositionRouteClearance */
    { (ASN_TAG_CLASS_CONTEXT | (84 << 2)), 84, 0, 0 }, /* uM84PositionProcedureName */
    { (ASN_TAG_CLASS_CONTEXT | (85 << 2)), 85, 0, 0 }, /* uM85RouteClearance */
    { (ASN_TAG_CLASS_CONTEXT | (86 << 2)), 86, 0, 0 }, /* uM86PositionRouteClearance */
    { (ASN_TAG_CLASS_CONTEXT | (87 << 2)), 87, 0, 0 }, /* uM87Position */
    { (ASN_TAG_CLASS_CONTEXT | (88 << 2)), 88, 0, 0 }, /* uM88PositionPosition */
    { (ASN_TAG_CLASS_CONTEXT | (89 << 2)), 89, 0, 0 }, /* uM89TimePosition */
    { (ASN_TAG_CLASS_CONTEXT | (90 << 2)), 90, 0, 0 }, /* uM90AltitudePosition */
    { (ASN_TAG_CLASS_CONTEXT | (91 << 2)), 91, 0, 0 }, /* uM91HoldClearance */
    { (ASN_TAG_CLASS_CONTEXT | (92 << 2)), 92, 0, 0 }, /* uM92PositionAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (93 << 2)), 93, 0, 0 }, /* uM93Time */
    { (ASN_TAG_CLASS_CONTEXT | (94 << 2)), 94, 0, 0 }, /* uM94DirectionDegrees */
    { (ASN_TAG_CLASS_CONTEXT | (95 << 2)), 95, 0, 0 }, /* uM95DirectionDegrees */
    { (ASN_TAG_CLASS_CONTEXT | (96 << 2)), 96, 0, 0 }, /* uM96NULL */
    { (ASN_TAG_CLASS_CONTEXT | (97 << 2)), 97, 0, 0 }, /* uM97PositionDegrees */
    { (ASN_TAG_CLASS_CONTEXT | (98 << 2)), 98, 0, 0 }, /* uM98DirectionDegrees */
    { (ASN_TAG_CLASS_CONTEXT | (99 << 2)), 99, 0, 0 }, /* uM99ProcedureName */
    { (ASN_TAG_CLASS_CONTEXT | (100 << 2)), 100, 0, 0 }, /* uM100TimeSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (101 << 2)), 101, 0, 0 }, /* uM101PositionSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (102 << 2)), 102, 0, 0 }, /* uM102AltitudeSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (103 << 2)), 103, 0, 0 }, /* uM103TimeSpeedSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (104 << 2)), 104, 0, 0 }, /* uM104PositionSpeedSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (105 << 2)), 105, 0, 0 }, /* uM105AltitudeSpeedSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (106 << 2)), 106, 0, 0 }, /* uM106Speed */
    { (ASN_TAG_CLASS_CONTEXT | (107 << 2)), 107, 0, 0 }, /* uM107NULL */
    { (ASN_TAG_CLASS_CONTEXT | (108 << 2)), 108, 0, 0 }, /* uM108Speed */
    { (ASN_TAG_CLASS_CONTEXT | (109 << 2)), 109, 0, 0 }, /* uM109Speed */
    { (ASN_TAG_CLASS_CONTEXT | (110 << 2)), 110, 0, 0 }, /* uM110SpeedSpeed */
    { (ASN_TAG_CLASS_CONTEXT | (111 << 2)), 111, 0, 0 }, /* uM111Speed */
    { (ASN_TAG_CLASS_CONTEXT | (112 << 2)), 112, 0, 0 }, /* uM112Speed */
    { (ASN_TAG_CLASS_CONTEXT | (113 << 2)), 113, 0, 0 }, /* uM113Speed */
    { (ASN_TAG_CLASS_CONTEXT | (114 << 2)), 114, 0, 0 }, /* uM114Speed */
    { (ASN_TAG_CLASS_CONTEXT | (115 << 2)), 115, 0, 0 }, /* uM115Speed */
    { (ASN_TAG_CLASS_CONTEXT | (116 << 2)), 116, 0, 0 }, /* uM116NULL */
    { (ASN_TAG_CLASS_CONTEXT | (117 << 2)), 117, 0, 0 }, /* uM117ICAOunitnameFrequency */
    { (ASN_TAG_CLASS_CONTEXT | (118 << 2)), 118, 0, 0 }, /* uM118PositionICAOunitnameFrequency */
    { (ASN_TAG_CLASS_CONTEXT | (119 << 2)), 119, 0, 0 }, /* uM119TimeICAOunitnameFrequency */
    { (ASN_TAG_CLASS_CONTEXT | (120 << 2)), 120, 0, 0 }, /* uM120ICAOunitnameFrequency */
    { (ASN_TAG_CLASS_CONTEXT | (121 << 2)), 121, 0, 0 }, /* uM121PositionICAOunitnameFrequency */
    { (ASN_TAG_CLASS_CONTEXT | (122 << 2)), 122, 0, 0 }, /* uM122TimeICAOunitnameFrequency */
    { (ASN_TAG_CLASS_CONTEXT | (123 << 2)), 123, 0, 0 }, /* uM123BeaconCode */
    { (ASN_TAG_CLASS_CONTEXT | (124 << 2)), 124, 0, 0 }, /* uM124NULL */
    { (ASN_TAG_CLASS_CONTEXT | (125 << 2)), 125, 0, 0 }, /* uM125NULL */
    { (ASN_TAG_CLASS_CONTEXT | (126 << 2)), 126, 0, 0 }, /* uM126NULL */
    { (ASN_TAG_CLASS_CONTEXT | (127 << 2)), 127, 0, 0 }, /* uM127NULL */
    { (ASN_TAG_CLASS_CONTEXT | (128 << 2)), 128, 0, 0 }, /* uM128Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (129 << 2)), 129, 0, 0 }, /* uM129Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (130 << 2)), 130, 0, 0 }, /* uM130Position */
    { (ASN_TAG_CLASS_CONTEXT | (131 << 2)), 131, 0, 0 }, /* uM131NULL */
    { (ASN_TAG_CLASS_CONTEXT | (132 << 2)), 132, 0, 0 }, /* uM132NULL */
    { (ASN_TAG_CLASS_CONTEXT | (133 << 2)), 133, 0, 0 }, /* uM133NULL */
    { (ASN_TAG_CLASS_CONTEXT | (134 << 2)), 134, 0, 0 }, /* uM134NULL */
    { (ASN_TAG_CLASS_CONTEXT | (135 << 2)), 135, 0, 0 }, /* uM135NULL */
    { (ASN_TAG_CLASS_CONTEXT | (136 << 2)), 136, 0, 0 }, /* uM136NULL */
    { (ASN_TAG_CLASS_CONTEXT | (137 << 2)), 137, 0, 0 }, /* uM137NULL */
    { (ASN_TAG_CLASS_CONTEXT | (138 << 2)), 138, 0, 0 }, /* uM138NULL */
    { (ASN_TAG_CLASS_CONTEXT | (139 << 2)), 139, 0, 0 }, /* uM139NULL */
    { (ASN_TAG_CLASS_CONTEXT | (140 << 2)), 140, 0, 0 }, /* uM140NULL */
    { (ASN_TAG_CLASS_CONTEXT | (141 << 2)), 141, 0, 0 }, /* uM141NULL */
    { (ASN_TAG_CLASS_CONTEXT | (142 << 2)), 142, 0, 0 }, /* uM142NULL */
    { (ASN_TAG_CLASS_CONTEXT | (143 << 2)), 143, 0, 0 }, /* uM143NULL */
    { (ASN_TAG_CLASS_CONTEXT | (144 << 2)), 144, 0, 0 }, /* uM144NULL */
    { (ASN_TAG_CLASS_CONTEXT | (145 << 2)), 145, 0, 0 }, /* uM145NULL */
    { (ASN_TAG_CLASS_CONTEXT | (146 << 2)), 146, 0, 0 }, /* uM146NULL */
    { (ASN_TAG_CLASS_CONTEXT | (147 << 2)), 147, 0, 0 }, /* uM147NULL */
    { (ASN_TAG_CLASS_CONTEXT | (148 << 2)), 148, 0, 0 }, /* uM148Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (149 << 2)), 149, 0, 0 }, /* uM149AltitudePosition */
    { (ASN_TAG_CLASS_CONTEXT | (150 << 2)), 150, 0, 0 }, /* uM150AltitudeTime */
    { (ASN_TAG_CLASS_CONTEXT | (151 << 2)), 151, 0, 0 }, /* uM151Speed */
    { (ASN_TAG_CLASS_CONTEXT | (152 << 2)), 152, 0, 0 }, /* uM152DistanceOffsetDirection */
    { (ASN_TAG_CLASS_CONTEXT | (153 << 2)), 153, 0, 0 }, /* uM153Altimeter */
    { (ASN_TAG_CLASS_CONTEXT | (154 << 2)), 154, 0, 0 }, /* uM154NULL */
    { (ASN_TAG_CLASS_CONTEXT | (155 << 2)), 155, 0, 0 }, /* uM155Position */
    { (ASN_TAG_CLASS_CONTEXT | (156 << 2)), 156, 0, 0 }, /* uM156NULL */
    { (ASN_TAG_CLASS_CONTEXT | (157 << 2)), 157, 0, 0 }, /* uM157Frequency */
    { (ASN_TAG_CLASS_CONTEXT | (158 << 2)), 158, 0, 0 }, /* uM158ATISCode */
    { (ASN_TAG_CLASS_CONTEXT | (159 << 2)), 159, 0, 0 }, /* uM159ErrorInformation */
    { (ASN_TAG_CLASS_CONTEXT | (160 << 2)), 160, 0, 0 }, /* uM160ICAOfacilitydesignation */
    { (ASN_TAG_CLASS_CONTEXT | (161 << 2)), 161, 0, 0 }, /* uM161NULL */
    { (ASN_TAG_CLASS_CONTEXT | (162 << 2)), 162, 0, 0 }, /* uM162NULL */
    { (ASN_TAG_CLASS_CONTEXT | (163 << 2)), 163, 0, 0 }, /* uM163ICAOfacilitydesignationTp4table */
    { (ASN_TAG_CLASS_CONTEXT | (164 << 2)), 164, 0, 0 }, /* uM164NULL */
    { (ASN_TAG_CLASS_CONTEXT | (165 << 2)), 165, 0, 0 }, /* uM165NULL */
    { (ASN_TAG_CLASS_CONTEXT | (166 << 2)), 166, 0, 0 }, /* uM166NULL */
    { (ASN_TAG_CLASS_CONTEXT | (167 << 2)), 167, 0, 0 }, /* uM167NULL */
    { (ASN_TAG_CLASS_CONTEXT | (168 << 2)), 168, 0, 0 }, /* uM168NULL */
    { (ASN_TAG_CLASS_CONTEXT | (169 << 2)), 169, 0, 0 }, /* uM169FreeText */
    { (ASN_TAG_CLASS_CONTEXT | (170 << 2)), 170, 0, 0 }, /* uM170FreeText */
    { (ASN_TAG_CLASS_CONTEXT | (171 << 2)), 171, 0, 0 }, /* uM171VerticalRate */
    { (ASN_TAG_CLASS_CONTEXT | (172 << 2)), 172, 0, 0 }, /* uM172VerticalRate */
    { (ASN_TAG_CLASS_CONTEXT | (173 << 2)), 173, 0, 0 }, /* uM173VerticalRate */
    { (ASN_TAG_CLASS_CONTEXT | (174 << 2)), 174, 0, 0 }, /* uM174VerticalRate */
    { (ASN_TAG_CLASS_CONTEXT | (175 << 2)), 175, 0, 0 }, /* uM175Altitude */
    { (ASN_TAG_CLASS_CONTEXT | (176 << 2)), 176, 0, 0 }, /* uM176NULL */
    { (ASN_TAG_CLASS_CONTEXT | (177 << 2)), 177, 0, 0 }, /* uM177NULL */
    { (ASN_TAG_CLASS_CONTEXT | (178 << 2)), 178, 0, 0 }, /* uM178NULL */
    { (ASN_TAG_CLASS_CONTEXT | (179 << 2)), 179, 0, 0 }, /* uM179NULL */
    { (ASN_TAG_CLASS_CONTEXT | (180 << 2)), 180, 0, 0 }, /* uM180AltitudeAltitude */
    { (ASN_TAG_CLASS_CONTEXT | (181 << 2)), 181, 0, 0 }, /* uM181ToFromPosition */
    { (ASN_TAG_CLASS_CONTEXT | (182 << 2)), 182, 0, 0 } /* uM182NULL */
};
static asn_CHOICE_specifics_t asn_SPC_FANSATCUplinkMsgElementId_specs_1 = {
	sizeof(struct FANSATCUplinkMsgElementId),
	offsetof(struct FANSATCUplinkMsgElementId, _asn_ctx),
	offsetof(struct FANSATCUplinkMsgElementId, present),
	sizeof(((struct FANSATCUplinkMsgElementId *)0)->present),
	asn_MAP_FANSATCUplinkMsgElementId_tag2el_1,
	183,	/* Count of tags in the map */
	0,
	-1	/* Extensions start */
};
asn_TYPE_descriptor_t asn_DEF_FANSATCUplinkMsgElementId = {
	"FANSATCUplinkMsgElementId",
	"FANSATCUplinkMsgElementId",
	CHOICE_free,
	CHOICE_print,
	CHOICE_constraint,
	CHOICE_decode_ber,
	CHOICE_encode_der,
	CHOICE_decode_xer,
	CHOICE_encode_xer,
	CHOICE_decode_uper,
	CHOICE_encode_uper,
	CHOICE_outmost_tag,
	0,	/* No effective tags (pointer) */
	0,	/* No effective tags (count) */
	0,	/* No tags (pointer) */
	0,	/* No tags (count) */
	&asn_PER_type_FANSATCUplinkMsgElementId_constr_1,
	asn_MBR_FANSATCUplinkMsgElementId_1,
	183,	/* Elements count */
	&asn_SPC_FANSATCUplinkMsgElementId_specs_1	/* Additional specs */
};

