#pragma once

#define FY3_A_SCID 49
#define FY3_B_SCID 50
#define FY3_C_SCID 51
#define FY3_D_SCID 52
#define FY3_E_SCID 56
#define FY3_F_SCID 60
#define FY3_G_SCID 64

#define FY3_A_NORAD 32958
#define FY3_B_NORAD 37214
#define FY3_C_NORAD 39260
#define FY3_D_NORAD 43010
#define FY3_E_NORAD 49008
#define FY3_F_NORAD 57490
#define FY3_G_NORAD 56232

#define FY3_ORBIT_HEIGHT 842

#define FY3_VIRR_SWATH 2800
#define FY3_VIRR_RES 1.1

#define FY3_MERSI_SWATH 2900
#define FY3_MERSILL_SWATH 2100 // Not sure what the swath is so... Tried what worked
#define FY3_MERSI_RES250 0.250
#define FY3_MERSI_RES1000 1