/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#include "FANSRouteClearance.h"

static asn_TYPE_member_t asn_MBR_FANSRouteClearance_1[] = {
	{ ATF_POINTER, 10, offsetof(struct FANSRouteClearance, airportDeparture),
		(ASN_TAG_CLASS_CONTEXT | (0 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAirportDeparture,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"airportDeparture"
		},
	{ ATF_POINTER, 9, offsetof(struct FANSRouteClearance, airportDestination),
		(ASN_TAG_CLASS_CONTEXT | (1 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAirportDestination,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"airportDestination"
		},
	{ ATF_POINTER, 8, offsetof(struct FANSRouteClearance, runwayDeparture),
		(ASN_TAG_CLASS_CONTEXT | (2 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSRunwayDeparture,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"runwayDeparture"
		},
	{ ATF_POINTER, 7, offsetof(struct FANSRouteClearance, procedureDeparture),
		(ASN_TAG_CLASS_CONTEXT | (3 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSProcedureDeparture,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"procedureDeparture"
		},
	{ ATF_POINTER, 6, offsetof(struct FANSRouteClearance, runwayArrival),
		(ASN_TAG_CLASS_CONTEXT | (4 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSRunwayArrival,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"runwayArrival"
		},
	{ ATF_POINTER, 5, offsetof(struct FANSRouteClearance, procedureApproach),
		(ASN_TAG_CLASS_CONTEXT | (5 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSProcedureApproach,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"procedureApproach"
		},
	{ ATF_POINTER, 4, offsetof(struct FANSRouteClearance, procedureArrival),
		(ASN_TAG_CLASS_CONTEXT | (6 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSProcedureArrival,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"procedureArrival"
		},
	{ ATF_POINTER, 3, offsetof(struct FANSRouteClearance, airwayIntercept),
		(ASN_TAG_CLASS_CONTEXT | (7 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSAirwayIntercept,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"airwayIntercept"
		},
	{ ATF_POINTER, 2, offsetof(struct FANSRouteClearance, routeinformation_seqOf),
		(ASN_TAG_CLASS_CONTEXT | (8 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSRouteInformationSequence,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"routeinformation-seqOf"
		},
	{ ATF_POINTER, 1, offsetof(struct FANSRouteClearance, routeInformationAdditional),
		(ASN_TAG_CLASS_CONTEXT | (9 << 2)),
		-1,	/* IMPLICIT tag at current level */
		&asn_DEF_FANSRouteInformationAdditional,
		0,	/* Defer constraints checking to the member type */
		0,	/* No PER visible constraints */
		0,
		"routeInformationAdditional"
		},
};
static const int asn_MAP_FANSRouteClearance_oms_1[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };
static const ber_tlv_tag_t asn_DEF_FANSRouteClearance_tags_1[] = {
	(ASN_TAG_CLASS_UNIVERSAL | (16 << 2))
};
static const asn_TYPE_tag2member_t asn_MAP_FANSRouteClearance_tag2el_1[] = {
    { (ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0 }, /* airportDeparture */
    { (ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0 }, /* airportDestination */
    { (ASN_TAG_CLASS_CONTEXT | (2 << 2)), 2, 0, 0 }, /* runwayDeparture */
    { (ASN_TAG_CLASS_CONTEXT | (3 << 2)), 3, 0, 0 }, /* procedureDeparture */
    { (ASN_TAG_CLASS_CONTEXT | (4 << 2)), 4, 0, 0 }, /* runwayArrival */
    { (ASN_TAG_CLASS_CONTEXT | (5 << 2)), 5, 0, 0 }, /* procedureApproach */
    { (ASN_TAG_CLASS_CONTEXT | (6 << 2)), 6, 0, 0 }, /* procedureArrival */
    { (ASN_TAG_CLASS_CONTEXT | (7 << 2)), 7, 0, 0 }, /* airwayIntercept */
    { (ASN_TAG_CLASS_CONTEXT | (8 << 2)), 8, 0, 0 }, /* routeinformation-seqOf */
    { (ASN_TAG_CLASS_CONTEXT | (9 << 2)), 9, 0, 0 } /* routeInformationAdditional */
};
static asn_SEQUENCE_specifics_t asn_SPC_FANSRouteClearance_specs_1 = {
	sizeof(struct FANSRouteClearance),
	offsetof(struct FANSRouteClearance, _asn_ctx),
	asn_MAP_FANSRouteClearance_tag2el_1,
	10,	/* Count of tags in the map */
	asn_MAP_FANSRouteClearance_oms_1,	/* Optional members */
	10, 0,	/* Root/Additions */
	-1,	/* Start extensions */
	-1	/* Stop extensions */
};
asn_TYPE_descriptor_t asn_DEF_FANSRouteClearance = {
	"FANSRouteClearance",
	"FANSRouteClearance",
	SEQUENCE_free,
	SEQUENCE_print,
	SEQUENCE_constraint,
	SEQUENCE_decode_ber,
	SEQUENCE_encode_der,
	SEQUENCE_decode_xer,
	SEQUENCE_encode_xer,
	SEQUENCE_decode_uper,
	SEQUENCE_encode_uper,
	0,	/* Use generic outmost tag fetcher */
	asn_DEF_FANSRouteClearance_tags_1,
	sizeof(asn_DEF_FANSRouteClearance_tags_1)
		/sizeof(asn_DEF_FANSRouteClearance_tags_1[0]), /* 1 */
	asn_DEF_FANSRouteClearance_tags_1,	/* Same as above */
	sizeof(asn_DEF_FANSRouteClearance_tags_1)
		/sizeof(asn_DEF_FANSRouteClearance_tags_1[0]), /* 1 */
	0,	/* No PER visible constraints */
	asn_MBR_FANSRouteClearance_1,
	10,	/* Elements count */
	&asn_SPC_FANSRouteClearance_specs_1	/* Additional specs */
};

