/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSErrorInformation_H_
#define	_FANSErrorInformation_H_


#include "asn_application.h"

/* Including external dependencies */
#include "NativeEnumerated.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Dependencies */
typedef enum FANSErrorInformation {
	FANSErrorInformation_applicationError	= 0,
	FANSErrorInformation_duplicateMsgIdentificationNumber	= 1,
	FANSErrorInformation_unrecognizedMsgReferenceNumber	= 2,
	FANSErrorInformation_endServiceWithPendingMsgs	= 3,
	FANSErrorInformation_endServiceWithNoValidResponse	= 4,
	FANSErrorInformation_insufficientMsgStorageCapacity	= 5,
	FANSErrorInformation_noAvailableMsgIdentificationNumber	= 6,
	FANSErrorInformation_commandedTermination	= 7,
	FANSErrorInformation_insufficientData	= 8,
	FANSErrorInformation_unexpectedData	= 9,
	FANSErrorInformation_invalidData	= 10,
	FANSErrorInformation_reservedErrorMsg1	= 11,
	FANSErrorInformation_reservedErrorMsg2	= 12,
	FANSErrorInformation_reservedErrorMsg3	= 13,
	FANSErrorInformation_reservedErrorMsg4	= 14,
	FANSErrorInformation_reservedErrorMsg5	= 15,
	FANSErrorInformation_reservedErrorMsg6	= 16
} e_FANSErrorInformation;

/* FANSErrorInformation */
typedef long	 FANSErrorInformation_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSErrorInformation;
asn_struct_free_f FANSErrorInformation_free;
asn_struct_print_f FANSErrorInformation_print;
asn_constr_check_f FANSErrorInformation_constraint;
ber_type_decoder_f FANSErrorInformation_decode_ber;
der_type_encoder_f FANSErrorInformation_encode_der;
xer_type_decoder_f FANSErrorInformation_decode_xer;
xer_type_encoder_f FANSErrorInformation_encode_xer;
per_type_decoder_f FANSErrorInformation_decode_uper;
per_type_encoder_f FANSErrorInformation_encode_uper;

#ifdef __cplusplus
}
#endif

#endif	/* _FANSErrorInformation_H_ */
#include "asn_internal.h"
