/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSRouteInformationAdditional_H_
#define	_FANSRouteInformationAdditional_H_


#include "asn_application.h"

/* Including external dependencies */
#include "constr_SEQUENCE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Forward declarations */
struct FANSATWAlongTrackWaypointSequence;
struct FANSReportingPoints;
struct FANSInterceptCourseFromSequence;
struct FANSHoldatwaypointSequence;
struct FANSWaypointSpeedAltitudesequence;
struct FANSRTARequiredTimeArrivalSequence;

/* FANSRouteInformationAdditional */
typedef struct FANSRouteInformationAdditional {
	struct FANSATWAlongTrackWaypointSequence	*aTWalongtrackwaypointsequence	/* OPTIONAL */;
	struct FANSReportingPoints	*reportingPoints	/* OPTIONAL */;
	struct FANSInterceptCourseFromSequence	*interceptCourseFromSequence	/* OPTIONAL */;
	struct FANSHoldatwaypointSequence	*holdAtWaypointSequence	/* OPTIONAL */;
	struct FANSWaypointSpeedAltitudesequence	*waypointSpeedAltitudeSequence	/* OPTIONAL */;
	struct FANSRTARequiredTimeArrivalSequence	*rTArequiredtimearrivalsequenc	/* OPTIONAL */;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} FANSRouteInformationAdditional_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSRouteInformationAdditional;

#ifdef __cplusplus
}
#endif

/* Referred external types */
#include "FANSATWAlongTrackWaypointSequence.h"
#include "FANSReportingPoints.h"
#include "FANSInterceptCourseFromSequence.h"
#include "FANSHoldatwaypointSequence.h"
#include "FANSWaypointSpeedAltitudesequence.h"
#include "FANSRTARequiredTimeArrivalSequence.h"

#endif	/* _FANSRouteInformationAdditional_H_ */
#include "asn_internal.h"
